"""
"""

import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

import numpy as np
import scipy.sparse as sparse
from scipy.sparse import coo_matrix, dia_matrix, csc_matrix, dok_matrix
import scipy.sparse as sparse
import pandas

import pylab as pl
from matplotlib.gridspec import GridSpec

from nacl.minimize import Minimizer, LogLikelihood2

from sksparse import cholmod

from nacl.lib import bspline
from nacl.lib.fitparameters import FitParameters
from nacl.loglikelihood import LogLikelihood

from matplotlib import cm

def gaussian(x, a, sigma):
    """
    Gaussian function *2
    """
    norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    return norm * np.exp(-0.5 * xx**2)

def gaussian_2d(x, y, a, sigma):
    """
    Gaussian function *2
    """
    norm = a / (np.sqrt(2. * np.pi) * sigma)
    xx = x / sigma
    yy = y / sigma
    return norm * np.exp(-0.5 * (xx**2 + yy**2))

def gen_data(xmin=-10., xmax=10., n=100, ):
    """
    Generates random data that follows a gaussian with noise
    """
    x = np.random.uniform(xmin, xmax, size=n)  #coordonnées
    err = np.linspace(0.1, 0.1, 20)
    noise = np.random.uniform(low=0.5, high=0.5, size=20)
    return x, err, gaussian(x) + noise


class ToyDataGenGenerator:

    def __init__(self, xmin=-10., xmax=10., yerr=0.01, a=1., sigma=3.,
                 n=100, delta=2., bsize=10, order=4):
        self.xmin, self.xmax = xmin, xmax
        self.yerr = yerr
        self.a, self.sigma = a, sigma
        self.n = n
        self.delta = delta
        self.basis = bspline.BSpline(np.linspace(xmin-delta, xmax+delta, bsize), order=order)

    def generate(self):
        """
        generate a realization of a dataset
        """
        x = np.random.uniform(self.xmin, self.xmax, size=self.n)
        y = gaussian(x, self.a, self.sigma)
        yerr = np.full(self.n, self.yerr)
        noise = np.random.normal(loc=0., scale=yerr)
        self.x, self.y_true, self.yerr, self.noise = x, y, yerr, noise
        self.y = self.y_true + self.noise
        return self.x, self.y + self.noise, self.yerr

    def get_data(self):
        if not hasattr(self, 'x'):
            self.generate()
        dd = pandas.DataFrame({'xx': self.x,
                               'y': self.y,
                               'yerr': self.yerr,
                               'bads': np.zeros(len(self.yerr)).astype(bool),
                               })
        return dd

    def fit(self, beta=1.E-6):
        """a fit of the data -- the simplest possible
        """
        J = self.basis.eval(self.x)
        w = 1. / self.yerr**2
        W = dia_matrix((w, [0]), shape=(self.n, self.n))
        H = J.T @ W @ J
        B = J.T @ W @ self.y
        fact = cholmod.cholesky(H, beta=beta)
        self.theta = fact(B)
    
    def plot(self, color='b', marker='.'):
        """plot the data and the model
        """
        fig = pl.figure()
        gs = GridSpec(2, 1, height_ratios=[4,2])

        ax1 = fig.add_subplot(gs[0])
        ax1.errorbar(self.x, self.y, yerr=self.yerr, ls='', color=color, marker=marker, label='data', zorder=0.5)
        ax1.set_ylabel('model \& data')
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        ax2.set_xlabel('x')
        ax2.set_ylabel('residuals')

        # pl.grid(ls=':')
        # pl.plot(self.basis.grid, np.zeros(len(self.basis.grid)), 'k|')
        if hasattr(self, 'theta'):
            xx = self.basis._grid.copy()
            JJ = self.basis.eval(xx)
            ax1.plot(xx, JJ@self.theta, ls='', color='red', marker='.', alpha=1., label='nodes', zorder=10., linewidth=2)

            xx = np.linspace(self.xmin-self.delta, self.xmax+self.delta, 200)
            JJ = self.basis.eval(xx)
            ax1.plot(xx, JJ@self.theta, 'r-', label='model', zorder=2.)

            JJ = self.basis.eval(self.x)
            ax2.errorbar(self.x, self.y-JJ @ self.theta, yerr=self.yerr, ls='', marker='.', color='k')

        pl.legend(loc='best')

class ToyDataGenGenerator2D:

    def __init__(self, x0_min=-10., x0_max=10., x1_min=-10., x1_max=10., yerr=0.01, a=1., sigma=3., n0=30, n1=30, delta=2., b0_size=10, b1_size=10, mu_0=0., mu_1=0., mu_max=0.):
        self.x0_min, self.x0_max = x0_min, x0_max
        self.x1_min, self.x1_max = x1_min, x1_max
        self.yerr = yerr
        self.a, self.sigma = a, sigma
        self.n0 = n0
        self.n1 = n1
        self.delta = delta
        self.basis = bspline.BSpline2D(np.linspace(x0_min-delta, x0_max+delta, b0_size), np.linspace(x1_min-delta, x1_max+delta, b1_size), x_order=4, y_order=4)
        self.mu = mu_max
        self.mu_0 = mu_0
        self.mu_1 = mu_1
        self.b0 = b0_size
        self.b1 = b1_size

    def generate(self):
        """
        generate a realization of a dataset
        """
        x0 = np.random.uniform(self.x0_min, self.x0_max, size=self.n0)
        x1 = np.random.uniform(self.x1_min, self.x1_max, size=self.n1)
        X0, X1 = np.meshgrid(x0, x1)
        X0 = X0.ravel()
        X1 = X1.ravel()
        n = len(X0)
        self.n = n
        yerr = np.full(self.n, self.yerr)
        y = gaussian_2d(X0, X1, self.a, self.sigma)
        noise = np.random.uniform(low = -self.yerr, high = self.yerr, size = n)
        self.x0, self.x1, self.y_true, self.noise, self.yerr = x0, x1, y, noise, yerr
        self.y = self.y_true + self.noise
        return self.x0, self.x1, self.y + self.noise, self.yerr
        
    def get_data(self):
        if not hasattr(self, 'x'):
            self.generate()
        X0, X1 = np.meshgrid(self.x0, self.x1)
        dd = pandas.DataFrame({'xx0': X0.ravel(),
                               'xx1': X1.ravel(),
                               'y': self.y,
                               'yerr': self.yerr,
                               'bads': np.zeros(len(self.yerr)).astype(bool),
                               })
        return dd
    
    def adapt_reg_matrix(self):
        """
        adding adapted regularization to test different methods
        """
        X0, X1 = np.meshgrid(self.x0, self.x1)
        J = self.basis.eval(X0.ravel(), X1.ravel())
        J = J.toarray()
        D = self.basis.hessian(x=X0.ravel(), y=X1.ravel())
        D = D[0] + D[1] + D[2]
        D.toarray()
        
        s_j =  np.sum(J, axis=0)
        s_j_tilde = np.sum(D, axis=0)
        s_j_tilde = np.array(s_j_tilde)[0]
        
        mu = np.ones(len(J[0])) * self.mu
        idx = s_j_tilde!=0
        #idx = np.array(idx)[0]
        int_mj = (mu - s_j)
        int_mj[idx] = int_mj[idx]/s_j_tilde[idx]
        #print(int_mj[idx])
        zeros = np.zeros(len(mu))
        
        mu_j = np.max([int_mj, zeros] , axis=0)
        
        M = dia_matrix((mu_j, [0]), shape=(len(mu), len(mu))).tocoo()
        return M
        
    def reg_matrix(self):
        """
        adding regularization
        """
        X0, X1 = np.meshgrid(self.x0, self.x1)
        J = self.basis.eval(X0.ravel(), X1.ravel())
        J = J.toarray()
        
        m_0 = np.ones(len(J[0])) * self.mu_0
        M_0 = dia_matrix( (m_0, [0]), shape=( len(m_0), len(m_0) ) ).tocoo()  #order 0
        
        m_1 = np.vstack( ([2.] * len(m_0), [-1.] * len(m_0), [-1.] * len(m_0), [-1.] * len(m_0), [-1.] * len(m_0)) )
        m_1[0, 0] = m_1[0, -1] = 3.
        M_1 = dia_matrix((m_1, [0, -1, 1, self.n0, -self.n0]), shape=(len(m_0),len(m_0))).tocoo()
        M_1 *= self.mu_1
        return M_0, M_1
    
    def fit(self):
        """a fit of the data -- the simplest possible
        """
        if self.mu == 0.:
            X0, X1 = np.meshgrid(self.x0, self.x1)
            J = self.basis.eval(X0.ravel(), X1.ravel())
            M0, M1 = self.reg_matrix()
            w = 1. / self.yerr**2
            W = dia_matrix((w, [0]), shape=(self.n, self.n))
            H = J.T @ W @ J + M0.T @ M0 + M1.T @ M1
            #H = J.T @ W @ J
            B = J.T @ W @ self.y
            fact = cholmod.cholesky(H, beta=1.E-6)
            self.theta = fact(B)
        else : 
            X0, X1 = np.meshgrid(self.x0, self.x1)
            J = self.basis.eval(X0.ravel(), X1.ravel())
            M = self.adapt_reg_matrix()
            D = self.basis.hessian(x=X0.ravel(), y=X1.ravel())
            D = D[0] + D[1] + D[2]
            w = 1. / self.yerr**2
            W = dia_matrix((w, [0]), shape=(self.n0, self.n0))
            H = J.T @ W @ J + (D @ M).T @ (D @ M)
            B = J.T @ W @ self.y
            fact = cholmod.cholesky(H, beta=1.E-6)
            self.theta = fact(B)
        resid = self.y-J @ self.theta
        errs = self.yerr
        return resid, errs
    
    def plot(self, color='b', marker='.'):
        """plot the data and the model
        """
        mu = self.mu
        fig = pl.figure()
        gs = GridSpec(2, 1, height_ratios=[7,2])

        ax1 = fig.add_subplot(gs[0], projection='3d')
        #ax1.errorbar(self.x, self.y, yerr=self.yerr, ls='', color=color, marker=marker, label='data', zorder=0.5)
        ax1.set_zlabel('model & data')
        ax1.set_xlabel('x0')
        ax1.set_ylabel('x1')
        ax2 = fig.add_subplot(gs[1], sharex=ax1)
        ax2.set_xlabel('x')
        ax2.set_ylabel('residuals')
        
        X0, X1 = np.meshgrid(self.x0, self.x1)
        ax1.errorbar(X0.ravel(), X1.ravel(), self.y, zerr = self.yerr, ecolor=('black',0.1), ls='', marker='.', color='blue')
        if self.mu ==0.:
            ax1.set_title('µ0 = ' +  str(self.mu_0) + ' , µ1 = ' + str(self.mu_1))
        else : 
            ax1.set_title('µ max = ' + str(self.mu))
            

        # pl.grid(ls=':')
        # pl.plot(self.basis.grid, np.zeros(len(self.basis.grid)), 'k|')
        
        resid = []
        errs = []
        if hasattr(self, 'theta'):
            #xx = self.basis._grid.copy()
            xx = self.basis.bx.grid
            yy = self.basis.by.grid
            XX, YY = np.meshgrid(xx, yy)
            JJ = self.basis.eval(XX.ravel(), YY.ravel())
            #ax1.plot(xx, yy, JJ@self.theta, ls='', color='red', marker='.', alpha=1., label='nodes', zorder=10., linewidth=2)
            fitting = JJ@self.theta
            fitting = np.reshape(fitting, (self.b0, self.b1))
            ax1.plot_surface(XX, YY, fitting, cmap=cm.coolwarm, label='nodes')

            #xx = np.linspace(self.x0_min-self.delta, self.x0_max+self.delta, 200)
            #JJ = self.basis.eval(xx)
            #ax1.plot(xx, JJ@self.theta, 'r-', label='model', zorder=2.)

            JJ = self.basis.eval(X0.ravel(), X1.ravel())
            x = np.linspace(self.x0_min, self.x0_max, len(X0.ravel()))
            ax2.errorbar(x, self.y-JJ @ self.theta, yerr=self.yerr, ls='', marker='.', color='k')
            
            residuals_data = self.y-JJ @ self.theta
            pl.figure()
            pl.hexbin(X0.ravel(), X1.ravel(), residuals_data/self.yerr, gridsize=15)
            pl.colorbar()
            pl.title('Weighted Residuals')
            
        pl.legend(loc='best')

def main_hand_made_fit(bsize=20, n=100, beta=1., order=4):
    gen = ToyDataGenGenerator(bsize=bsize, n=n, order=order)
    r = gen.generate()
    gen.fit()
    gen.plot()

def main_hand_made_fit_2D(bsize=20, mu_max=0., mu_0=50., mu_1=50., plot=True, n0=30, n1=30):
    gen = ToyDataGenGenerator2D(b0_size=bsize, b1_size=bsize, n0=n0, n1=n1, mu_max=mu_max, mu_0=mu_0, mu_1=mu_1)
    r = gen.generate()
    resid, err = gen.fit()
    if plot:
        gen.plot()
    return resid, err

def check_grad(model, p):
    """
    """
    v,jacobian = model(p, jac=True)
    dx = 1.E-7
    pars = model.pars.copy()
    pars.free = p
    df = []
    for i in range(len(pars.full)):
        k = pars.indexof(i)
        if k < 0:
            continue
        pars.full[i] += dx
        vp = model(pars.free, jac=False)
        df.append((vp-v)/dx)
        pars.full[i] -= dx
    return np.array(jacobian.todense()), np.vstack(df).T


def check_deriv(pen, p, dx=1.E-6):
    """
    """
    v, grad, hess = pen(p=p, deriv=True)
    p0 = p.copy()

    df, d2f = [], []
    for i in range(len(p0)):
        p0[i] += dx
        vp = pen(p0, deriv=False)
        p0[i] -= (2*dx)
        vm = pen(p0, deriv=False)
        df.append((vp-vm)/(2*dx))
        p0[i] += dx
    return np.array(grad), np.vstack(df).T


class Model:
    """A spline description of the data

    Encpsulating the sline evaluation in a model, like that
    allows to easily fix parameters, use the nacl minimizer,
    with regularization penalties.
    """
    def __init__(self, x, grid=None, order=4):
        self.x = x
        self.grid = grid
        self.basis = bspline.BSpline(grid, order=4)
        self.J = self.basis.eval(self.x)
        self.pars = FitParameters([('theta', len(self.basis))])

    def __call__(self, p=None, jac=False):
        if p is not None:
            self.pars.free = p

        vals = self.J @ self.pars['theta'].full
        if not jac:
            return vals

        j = self.pars.indexof(self.J.col)
        idx = j>=0

        N = self.J.shape[0]
        n_free_pars = len(self.pars.free)
        JJ = coo_matrix((self.J.data[idx], (self.J.row[idx], j[idx])), shape=(N,n_free_pars))
        return vals, JJ

class Model_2D:
    """A spline description of the data in 2D

    Encpsulating the sline evaluation in a 2D model, like that
    allows to easily fix parameters, use the nacl minimizer,
    with regularization penalties.
    """
    def __init__(self, x0, x1, grid_x0=None, grid_x1=None, order=4):
        self.x0 = x0
        self.x1 = x1
        self.grid_x0 = grid_x0
        self.grid_x1 = grid_x1
        self.basis = bspline.BSpline2D(grid_x0, grid_x1, x_order=order, y_order=order)
        X0, X1 = np.meshgrid(x0, x1)
        self.J = self.basis.eval(X0.ravel(), X1.ravel())
        self.pars = FitParameters([('theta', len(self.basis))])

    def __call__(self, p=None, jac=False):
        if p is not None:
            self.pars.free = p

        vals = self.J @ self.pars['theta'].full
        if not jac:
            return vals

        j = self.pars.indexof(self.J.col)
        idx = j>=0

        N = self.J.shape[0]
        n_free_pars = len(self.pars.free)
        JJ = coo_matrix((self.J.data[idx], (self.J.row[idx], j[idx])), shape=(N,n_free_pars))
        return vals, JJ


class Regularization:
    """Regularization penalty

    TODO: it would be a good idea if the adaptive regularization was performed
    here. Or at least if there was a method to do it on the fly here.
    """
    def __init__(self, pars, block_name='theta', mu0=1.E-6, mu1=1.E-3):
        """
        """
        # should be a direct handle to the parameters
        # otherwise, if we decide to fix the model,
        # this will not be propagated
        self.pars = pars   # .copy()

        self.block_name = block_name
        self.mu0 = mu0
        self.mu1 = mu1
        self.npars = len(self.pars.full)
        self.matrix = self.init_reg_matrix()

    def init_reg_matrix(self):
        """
        """
        pars_full = self.pars.copy()
        pars_full.release()

        n = len(pars_full[self.block_name].full)
        N = len(pars_full.full)

        # order 0
        P = dia_matrix((np.ones(N), [0]), shape=(n, n)).tocoo()
        i = pars_full.indexof(P.row)
        j = pars_full.indexof(P.col)
        P = coo_matrix((P.data, (i,j)), shape=(N,N))

        # order 1
        # data = -np.ones((N,3))
        # data[:,2] = data[:,0] = 1.
        # data[1:-1,1] = -2.
        #
        # second order penality, from the laplacian matrix and slightly hacked
        # to make the end of the diagonal similar to the beginning
        data = np.ones((n,3))
        data[:,1] = -2
        Q = dia_matrix((data.T, [0,1,2]), shape=[n,n]).tocoo()
        # print(Q.todense())
        i = pars_full.indexof(Q.row)
        j = pars_full.indexof(Q.col)
        Q = coo_matrix((Q.data, (i,j)), shape=(N,N))
        A = Q.T @ Q
        B = A[:3,:3]
        A[-3:,-3:] = B[::-1,::-1]

        return self.mu0 * P + self.mu1 * A

    def ndof(self):
        return -len(self.pars.free)

    def get_log(self):
        return {}

    def __call__(self, p=None, deriv=False):
        """
        """
        if p is not None:
            self.pars.free = p

        pp = self.pars.full
        penalty = pp.T @ self.matrix @ pp

        if not deriv:
            return penalty

        n = self.matrix.shape[0]
        idx = self.pars.indexof(np.arange(n)) >= 0

        grad = 2. * (self.matrix @ pp)[idx]
        hess = 2. * self.matrix[:,idx][idx,:]

        return penalty, grad, hess


class SimplePedestalErrorModel:
    """Error snake variance

    A simple model: just a constant pedestal error added to the data
    """
    def __init__(self, model):
        """constructor
        """
        self.model = model

    def get_struct(self):
        """structure of the error model specific parameters
        """
        return [("gamma", 1)]

    def __call__(self, pars, jac=False):
        """
        """
        v, J = None, None
        if jac:
            v, J = self.model(pars, jac=True)
        else:
            v = self.model(pars, jac=False)
        g = pars['gamma'].full[0]

        var = g**2

        if not jac:
            return var

        N = len(v)
        n_free_pars = len(pars.free)
        dvar = np.full(2. * g)
        i = np.arange(N)
        j = pars['gamma'].indexof(np.zeros(N))
        idx = j>=0
        J = coo_matrix((dvar[idx], (i[idx], j[idx])), shape=(N,n_free_pars))

        return var, J


class ConstantErrorSnake:
    """Simplistic Error snake

    This model is slightly more complicated: the variance depends on the model value:

    ..math::
        \sigma = \gamma \times f(\lambda,p)
    """
    def __init__(self, model):
        pass

    def get_struct(self):
        return [('gamma', 1)]

    def __call__(self, p, jac=False):
        pass


class VariableErrorSnake:
    """Realistic Error Snake

    This model is even more complicated: the variance depends on the model value
    and varies as a function of math:`$x$`

    """
    def __init__(self, model, basis=None):
        pass

    def get_struct(self):
        pass

    def __call__(self, p, jac=False):
        pass



class CalibErrorModel:

    def __init__(self):
        pass

    def __call__(self, p, jac=False):
        pass


def plot_model_and_data(model, data, title='', reg_lambda=None):
    fig = pl.figure()
    gs = GridSpec(3, 1, height_ratios=[4,2,2])

    ax1 = fig.add_subplot(gs[0])
    ax1.errorbar(data.xx, data.y, yerr=data.yerr, ls='', marker='.')
    xx = np.linspace(model.basis.grid.min(), model.basis.grid.max(), 200)
    J = model.basis.eval(xx)
    ax1.plot(xx, J @ model.pars['theta'].full, 'r-', zorder=10)
    pl.setp(ax1.get_xticklabels(), visible=False)

    ax1.set_ylabel('model \& data')
    ax2 = fig.add_subplot(gs[1], sharex=ax1)
    res = data.y-model(model.pars.free)
    ax2.errorbar(data.xx, res, yerr=data.yerr, ls='', marker='.')
    ax2.set_xlabel('x')
    ax2.set_ylabel('residuals')
    pl.setp(ax2.get_xticklabels(), visible=False)

    ax3 = fig.add_subplot(gs[2], sharex=ax1)
    pars = model.pars.copy()
    xx = np.linspace(model.basis.grid.min(), model.basis.grid.max(), 500)
    J = model.basis.eval(xx)
    for i in range(len(pars.full)):
        pars.full[:] = 0.
        pars.full[i] = 3 * res.std()
        if reg_lambda is not None:
            color = pl.cm.jet(int(reg_lambda[i]/reg_lambda.max() * 256))
            # color = 'r' if reg_lambda[i]>0. else 'b'
        else:
            color = pl.cm.jet(int(i * 256/len(pars.full)))
        ax3.plot(xx, J@pars.full, ls='-', color=color)
    ax3.set_xlabel('x')

    pl.subplots_adjust(hspace=0.05)
    fig.suptitle(title)
    
def plot_model_and_data_2D(model, data, title='', reg_lambda=None):
    fig = pl.figure()
    gs = GridSpec(3, 1, height_ratios=[4,2,2])

    ax1 = fig.add_subplot(gs[0], projection='3d')
    #X0, X1 = np.meshgrid(data.xx0, data.xx1)
    ax1.errorbar(data.xx0, data.xx1, data.y, zerr=data.yerr, ecolor=('black',0.1), ls='', marker='.', color='blue')
    
    #ax1.errorbar(data.xx, data.y, yerr=data.yerr, ls='', marker='.')
    xx0 = np.linspace(model.basis.bx.grid.min(), model.basis.bx.grid.max(), 200)
    xx1 = np.linspace(model.basis.by.grid.min(), model.basis.by.grid.max(), 200)
    
    XX0, XX1 = np.meshgrid(xx0, xx1)
    
    J = model.basis.eval(XX0.ravel(), XX1.ravel())
    #ax1.plot(xx, J @ model.pars['theta'].full, 'r-', zorder=10)
    YY = J@model.pars['theta'].full
    YY = np.reshape(YY, (200,200))
    ax1.plot_surface(XX0, XX1, YY, cmap=cm.coolwarm)
    # pl.setp(ax1.get_xticklabels(), ax1.get_yticklabels(), visible=False) ??

    ax1.set_zlabel('model \& data')
    ax2 = fig.add_subplot(gs[1], sharex=ax1)
    res = data.y-model(model.pars.free)
    ax2.errorbar(data.xx0, res, yerr=data.yerr, ls='', marker='.')
    ax2.set_xlabel('x')
    ax2.set_ylabel('residuals')
    #pl.setp(ax2.get_xticklabels(), visible=False)

    #ax3 = fig.add_subplot(gs[2], sharex=ax1)
    #pars = model.pars.copy()
    #xx = np.linspace(model.basis.bx.grid.min(), model.basis.bx.grid.max(), 500)
    #J = model.basis.eval(xx)
    #for i in range(len(pars.full)):
    #    pars.full[:] = 0.
    #    pars.full[i] = 3 * res.std()
    #    if reg_lambda is not None:
    #        color = pl.cm.jet(int(reg_lambda[i]/reg_lambda.max() * 256))
    #        # color = 'r' if reg_lambda[i]>0. else 'b'
    #    else:
    #        color = pl.cm.jet(int(i * 256/len(pars.full)))
    #    ax3.plot(xx, J@pars.full, ls='-', color=color)
    #ax3.set_xlabel('x')

    pl.subplots_adjust(hspace=0.05)
    fig.suptitle(title)


def main(bsize=10, n=100, mu_reg=10., mu_reg_model=1., mu0=1.E-6, mu1=1.E-3, delta=10., order=4, fix=False,
         adapted_fit_1=False, adapted_fit_2=False, forced=False):
    """
    """
    gen = ToyDataGenGenerator(bsize=bsize, n=n, delta=delta, order=order)
    gen.generate()

    model = Model(gen.x, gen.basis.grid)
    n = len(model.pars['theta'].full)
    if fix:
        for i in range(6):
            model.pars['theta'].fix(i)
            model.pars['theta'].fix(n-i-1)

    reg = Regularization(pars = model.pars, mu0=mu0, mu1=mu1)
    # error_snake = ErrorSnake()
    # calib_error_model = CalibErrorModel()

    logging.info('FIRST FIT')
    ll = LogLikelihood(model, data=gen.get_data(), reg=[reg]) # , reg=reg, variance_model=error_snake)
    minz = Minimizer(ll)
    n_free_pars = len(model.pars.free)
    ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
    # ret = minz.minimize(p_init=np.random.rand(n_free_pars), beta=1.E-6)
    plot_model_and_data(model, gen.get_data(), 'raw')

    # iterate with adapted regularization
    if adapted_fit_1:
        logging.info('ADAPTED FIT')
        model.pars.release()
        reg = Regularization(pars=model.pars, mu0=mu0, mu1=mu1)
        _, J = model(model.pars.free, jac=1)
        l = np.array(np.maximum(mu_reg - J.sum(axis=0), 0.) / reg.matrix.sum(axis=0)).squeeze()
        L = dia_matrix((l, [0]), shape=reg.matrix.shape)
        pl.figure()
        pl.plot(l, 'k.')
        reg.matrix = reg.matrix @ L
        ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        minz = Minimizer(ll)
        ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        plot_model_and_data(model, gen.get_data(), title='adapted')
        pl.figure()
        pl.imshow(reg.matrix.todense())
        pl.title('adapted')

    if adapted_fit_2:
        logging.info('ADAPTED FIT 2')
        model.pars.release()
        reg = Regularization(pars=model.pars, mu0=2., mu1=1.)
        _, J = model(model.pars.free, jac=1)
        # l = np.array(np.maximum(mu_reg - J.sum(axis=0), 0.) / reg.matrix.sum(axis=0)).squeeze()
        # b = np.bincount(J.col, weights=J.data, minlength=J.shape[1])
        # bb = np.maximum(np.bincount(J.col, minlength=J.shape[1]), 1)
        # b /= bb
        # l = np.zeros_like(b)
        # l[b<0.2] = mu_reg

        cols = np.arange(J.shape[1])
        bins = np.arange(-0.5, J.shape[1]+0.5, 1.)
        d = np.digitize(J.col, bins, right=False) - 1
        l = np.array([0. if (d==col).sum() == 0 else J.data[d==col].max() for col in cols])
        idx = l<0.5
        l[idx] = mu_reg
        l[~idx] = mu_reg_model
        print(d)
        print(l)

        L = dia_matrix((l, [0]), shape=reg.matrix.shape)
        reg.matrix = reg.matrix @ L

        # model.pars.release()
        # nn = len(model.pars.full)
        # model.pars.fix(nn-1)
        # model.pars.fix(0)

        pl.figure()
        pl.plot(l, 'b')
        pl.title('$\Lambda$ (adapted regularization 2)')
        ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        minz = Minimizer(ll)
        n_free_pars = len(model.pars.free)
        ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        plot_model_and_data(model, gen.get_data(), title='adapted 2', reg_lambda=l)
        pl.figure()
        pl.imshow(reg.matrix.todense())
        pl.title('adapted 2')


        if forced:
            logging.info('FORCED')
            model.pars.full[:7] = 0.
            model.pars.full[-8:] = 0.
            plot_model_and_data(model, gen.get_data(), title='forced')
            v = ll(model.pars.free)
            msg = f'chi2={ll.log["main_chi2"]:.6e} | log_det_v={ll.log["log_det_v"]} | cons='
            msg += f'{ll.log["cons"]:8.6e}'
            msg += ' | reg='
            msg += f'{ll.log["reg"]:.6e}'
            logging.info(msg)

        # logging.info('FIT FORCED')
        # nn = len(model.pars.full)
        # for i in range(8):
        #     model.pars.fix(i)
        #     model.pars.fix(nn-i-1)
        # n_free_pars = len(model.pars.free)
        # ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        # minz = Minimizer(ll)
        # ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        # plot_model_and_data(model, gen.get_data(), title='forced fit')



    return model, gen.get_data(), minz, ret

def main_2D(bsize=10, n0=40, n1=50, mu_reg=10., mu_reg_model=1., mu0=1.E-6, mu1=1.E-3, delta=10., order=4, fix=False,
         adapted_fit_1=False, adapted_fit_2=False, forced=False):
    """
    """
    n=n0*n1
    gen = ToyDataGenGenerator2D(b0_size=bsize, b1_size=bsize, n0=n0, n1=n1)
    gen.generate()

    model = Model_2D(gen.x0, gen.x1, gen.basis.bx.grid, gen.basis.by.grid)
    n = len(model.pars['theta'].full)
    if fix:
        #for i in range(6):
        #    model.pars['theta'].fix(i)
        #    model.pars['theta'].fix(n-i-1)
        pass

    reg = Regularization(pars = model.pars, mu0=mu0, mu1=mu1)
    # error_snake = ErrorSnake()
    # calib_error_model = CalibErrorModel()

    logging.info('FIRST FIT')
    ll = LogLikelihood(model, data=gen.get_data(), reg=[reg]) # , reg=reg, variance_model=error_snake)
    minz = Minimizer(ll)
    n_free_pars = len(model.pars.free)
    ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
    # ret = minz.minimize(p_init=np.random.rand(n_free_pars), beta=1.E-6)
    plot_model_and_data_2D(model, gen.get_data(), 'raw')

    # iterate with adapted regularization
    if adapted_fit_1:
        logging.info('ADAPTED FIT')
        model.pars.release()
        reg = Regularization(pars=model.pars, mu0=mu0, mu1=mu1)
        _, J = model(model.pars.free, jac=1)
        l = np.array(np.maximum(mu_reg - J.sum(axis=0), 0.) / reg.matrix.sum(axis=0)).squeeze()
        L = dia_matrix((l, [0]), shape=reg.matrix.shape)
        pl.figure()
        pl.plot(l, 'k.')
        reg.matrix = reg.matrix @ L
        ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        minz = Minimizer(ll)
        ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        plot_model_and_data(model, gen.get_data(), title='adapted')
        pl.figure()
        pl.imshow(reg.matrix.todense())
        pl.title('adapted')

    if adapted_fit_2:
        logging.info('ADAPTED FIT 2')
        model.pars.release()
        reg = Regularization(pars=model.pars, mu0=2., mu1=1.)
        _, J = model(model.pars.free, jac=1)
        # l = np.array(np.maximum(mu_reg - J.sum(axis=0), 0.) / reg.matrix.sum(axis=0)).squeeze()
        # b = np.bincount(J.col, weights=J.data, minlength=J.shape[1])
        # bb = np.maximum(np.bincount(J.col, minlength=J.shape[1]), 1)
        # b /= bb
        # l = np.zeros_like(b)
        # l[b<0.2] = mu_reg

        cols = np.arange(J.shape[1])
        bins = np.arange(-0.5, J.shape[1]+0.5, 1.)
        d = np.digitize(J.col, bins, right=False) - 1
        l = np.array([0. if (d==col).sum() == 0 else J.data[d==col].max() for col in cols])
        idx = l<0.5
        l[idx] = mu_reg
        l[~idx] = mu_reg_model
        print(d)
        print(l)

        L = dia_matrix((l, [0]), shape=reg.matrix.shape)
        reg.matrix = reg.matrix @ L

        # model.pars.release()
        # nn = len(model.pars.full)
        # model.pars.fix(nn-1)
        # model.pars.fix(0)

        pl.figure()
        pl.plot(l, 'b')
        pl.title('$\Lambda$ (adapted regularization 2)')
        ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        minz = Minimizer(ll)
        n_free_pars = len(model.pars.free)
        ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        plot_model_and_data(model, gen.get_data(), title='adapted 2', reg_lambda=l)
        pl.figure()
        pl.imshow(reg.matrix.todense())
        pl.title('adapted 2')


        if forced:
            logging.info('FORCED')
            model.pars.full[:7] = 0.
            model.pars.full[-8:] = 0.
            plot_model_and_data(model, gen.get_data(), title='forced')
            v = ll(model.pars.free)
            msg = f'chi2={ll.log["main_chi2"]:.6e} | log_det_v={ll.log["log_det_v"]} | cons='
            msg += f'{ll.log["cons"]:8.6e}'
            msg += ' | reg='
            msg += f'{ll.log["reg"]:.6e}'
            logging.info(msg)

        # logging.info('FIT FORCED')
        # nn = len(model.pars.full)
        # for i in range(8):
        #     model.pars.fix(i)
        #     model.pars.fix(nn-i-1)
        # n_free_pars = len(model.pars.free)
        # ll = LogLikelihood(model, data=gen.get_data(), reg=[reg])
        # minz = Minimizer(ll)
        # ret = minz.minimize_lm(p_init=np.random.rand(n_free_pars), lamb=1.E-6)
        # plot_model_and_data(model, gen.get_data(), title='forced fit')



    return model, gen.get_data(), minz, ret

# def main(bsize=20, n=100):
#     gen = ToyDataGenGenerator(bsize=bsize, n=n)
#     r = gen.generate()
#     model = Model(gen.x, gen.y, gen.basis.grid)

#     ll = LogLikelihood2(model)
#     minz = Minimizer(ll)

#     return minz




# def get_naked_regularization_matrix(size, order=0):
#     """
#     Regularization matrix
#     """
#     if order == 0:
#         data = np.ones(size)
#         return dia_matrix((data, [0]), shape=(size,size)).tocoo()

#     data = np.vstack(([3.] * size, [-1.] * size, [-1.] * size))
#     data[0, 0] = data[0, -1] = 2.
#     mat = dia_matrix((data, [0, -1, 1]), shape=(size,size)).tocoo()
#     return mat


# class Model:
#     """
#     Toy model evaluated on a 1D spline basis
#     """
#     def __init__(self, x_range=(-10., 10.), basis_knots=10, x_data=None):
#         self.basis = self.init_basis(x_range=x_range, basis_knots=basis_knots)
#         self.init_pars()
#         self.x_data=x_data

#     def init_pars(self):
#         d = [('M', self.basis.nj)]
#         fp = FitParameters(list(set(d)))
#         fp['M'].full[:] = 1.
#         self.pars = fp

#     def init_basis(self, x_range=(-10., 10.), basis_knots=20):
#         x_grid = np.linspace(x_range[0], x_range[1], basis_knots)
#         basis = bspline.BSpline(x_grid, order=4)
#         return basis

#     def evals(self):
#         m = self.pars['M'].full
#         j = self.basis.eval(self.x_data)
#         #return gaussian(j@x)
#         return j@m

#     def jac(self):
#         m = self.pars['M'].full  #coeffs
#         j = self.basis.eval(self.x_data)   #J
#         M = j@m
#         n = len(m)  #len of coeffs
#         jj = []  #jacobian of the model
#         j_ = j.toarray()
#         for i in range(n):
#             #yy = -2*X*gaussian(X)
#             yy = -2*self.x_data*M
#             yy = j_[:,i]*yy
#             jj.append(yy)
#         jj = np.array(jj)
#         jj=jj.T
#         return csc_matrix(j)

#     def __call__(self, pars, deriv=True):
#         if deriv:
#             v = self.evals()
#             j = self.jac()
#             return v, j
#         else:
#             v = self.evals()
#             return v

# def chi2(model, Y, W, mat):
#     """
#     work in progress, minimizing chi2 with reg matrix
#     """
#     v, jac = model(model.pars.full, deriv=True)
#     pars = model.pars.full
#     R = Y - v
#     chi2_l = R.T@W@R
#     chi2_reg = pars.T@mat@pars
#     return chi2_l + chi2_reg


# def fit():
#     """
#     Fitting the surface on randomly generated data that follow a gaussian
#     """
#     x, err, f = gen_data()   #x, err on f, f
#     model = Model(x_data=x)
#     basis = model.basis

#     mat = get_naked_regularization_matrix(size=model.basis.nj, order=0)  #hasnt been used yet

#     #shouldn't be necessary but im projecting the generated surface on the same basis of the model
#     J = basis.eval(x)
#     W = dia_matrix((1./err**2, 0), shape=(len(x), len(x)))
#     H = J.T @ W @ J
#     factor = cholmod.cholesky(H, beta=1e-3)
#     Y = J@factor(J.T@W@f)

#     #initialising chi2
#     chi2_ndof = 2.
#     chi2_old = 10.

#     chi2_t = []
#     r_average = []  #average residuals
#     n=0  #number of iterations

#     #for i in range(10):
#     while np.abs(chi2_old - chi2_ndof) >0.001:
#         chi2_old = chi2_ndof

#         v, jac = model(model.pars.full, deriv=True)
#         R = Y - v
#         h = jac.T@W@jac
#         ff = cholmod.cholesky(h, beta=1e-3)
#         dtheta = ff(jac.T@W@R)
#         model.pars.full+=dtheta
#         n+=1
#         #chi2_ndof = np.sum(R**2/err**2)

#         #smth along the line of M.T@mat@M added to chi2

#         chi2_ndof = R.T@W@R
#         chi2_t.append(chi2_ndof)
#         r_average.append(np.sum(R)/len(R))
#     v_final, jj = model(model.pars.full)

#     resid = Y - v_final
#     return resid, model, x, f, err, n, chi2_t, r_average   #return final residuals, the model after fit, the generated data: x; f; err_f, n iteration, chi2, r_average

# def plot(r):
#     """
#     Plotter. Here 'r' is the result of the fit (so r=fit() ).
#     """
#     m = r[1]
#     pl.figure()
#     pl.errorbar(r[2], r[3], r[4], fmt='.', label='generated data')
#     pl.plot(r[2], m(m.pars.full,deriv=False), 'r.', label = 'fit')
#     pl.xlabel('x')
#     pl.ylabel('y')
#     pl.title('Toy Model Gaussian fit')
#     pl.legend()
#     pl.show()

#     pl.figure()
#     pl.plot(r[-2], 'k.')
#     pl.xlabel('iterations')
#     pl.ylabel('chi2')
#     pl.title('total chi2')
#     pl.show()

#     pl.figure()
#     pl.plot(r[2], r[0]/r[4], 'b.')
#     pl.xlabel('x')
#     pl.ylabel('R/err')
#     pl.title('weighted residuals')
#     pl.show()

#     #model after fit
#     J = m.basis.eval(m.basis.grid)
#     M = J@m.pars['M'].full

#     pl.figure()
#     pl.plot(m.basis.grid, M, 'g.')
#     pl.xlabel('x')
#     pl.ylabel('y')
#     pl.title('model')
#     pl.show()



import os
import os.path as op
import numpy as np
import scipy.sparse
import saunerie
import croaks
from saunerie.fitparameters import FitParameters

from salt3.minimizers import minimize_nr_var_quad
import salt3.generator
import salt3.models
import salt3.plots

from helpers import locate_test_datafile
from helpers import check_grad


class ToySplineModel:
    """A simple Toy spline model of the form: 
    
    just one flux per SN and a spline template common to all SNe. 
    No tmax, no stretch. 

    This model is intrinsically degenerate and should be fitted with constraints. 
    
    Simple, sparse constraints such as:
     - S(phase=0) = 1.
     - S'(phase=0) = 0
     generally work well
    """
    def __init__(self, dp, phase_range=[-100., 100.], grid_size=30):
        self.dp = dp
        self.nb_sn = len(dp.sn_set)
        self.spline_basis = saunerie.bspline.CardinalBSplineC(grid_size, phase_range)
        self.n_theta = grid_size
        self.pars = self.init_pars(dp)
        phases = self.dp.date
        self.jacobian = self.spline_basis.eval(phases)
        
    def init_pars(self, dp):
        fp = FitParameters([('flux', self.nb_sn), ("theta", len(self.spline_basis))])
        fp['flux'].full[:] = 1.
        fp['theta'].full[:] = 1.E-6
        return fp

    def __call__(self, p, jac=False, dp=None):
        """
        """
        self.pars.free = p
        if dp is None:
            dp = self.dp
            jacobian = self.jacobian
        else:
            phases = dp.date
            jacobian = self.spline_basis.eval(phases)
            
        sn_index = dp.sn_index
        flx = self.pars['flux'].full[sn_index]
        
        v = flx * (jacobian @ self.pars['theta'].full)

        if jac:
            N = len(dp.nt)
            n = len(self.spline_basis)
            i = np.arange(N)
            
            # dmodel_dflux 
            i_dm_dflux = i
            j_dm_dflux = self.pars['flux'].indexof(sn_index)
            v_dm_dflux = jacobian @ self.pars['theta'].full

            # dmodel_dtheta
            i_dm_dtheta = jacobian.row
            j_dm_dtheta = self.pars['theta'].indexof(jacobian.col)
            v_dm_dtheta = self.pars['flux'].full[sn_index[jacobian.row]] * jacobian.data
            
            i = np.hstack((i_dm_dflux, i_dm_dtheta))
            j = np.hstack((j_dm_dflux, j_dm_dtheta))            
            val = np.hstack((v_dm_dflux, v_dm_dtheta))
            
            idx_fix = np.where(j != -1)[0]
            DZ = scipy.sparse.coo_matrix((val[idx_fix], (i[idx_fix], j[idx_fix])), shape=(N, len(self.pars.free)))
            
            return v, DZ
        
        return v


class ToySplineNonIntegralConstraints():
    """
    .. note:: this implementation of the constraints 
              does adapt the size of the H matrix 
              to the number of fixed parameters. 
              On the other hand, it does not eliminate
              the constraints that have become irrelevant. 
              Then, the H.T matrix which is returned by __call__() 
              may be of rank < 3
    """
    def __init__(self, pars, basis, vals=[1., 0.]):
        self.pars = pars
        self.basis = basis
        self.vals = self.d = np.array(vals)
        assert(len(self.d) == 2)
        
        n = len(self.pars.full)
        
        # first constraint: S(0) = 1
        jacobian = self.basis.eval(np.array([0]))
        i1 = pars['theta'].indexof(jacobian.col)
        j1 = np.full(len(jacobian.col), 0)
        v1 = jacobian.data.copy()
        # second constraint: S'(0) = 0
        Jp = self.basis.deriv(np.array([0]))
        i2 = pars['theta'].indexof(Jp.col)
        j2 = np.full(len(Jp.col), 1)
        v2 = Jp.data.copy()
        self.H = scipy.sparse.coo_matrix((np.hstack((v1,v2)), (np.hstack((i1,i2)), np.hstack((j1,j2)))), shape=(n,2)).tocsr()
        
    # def __len__(self):
    #     return 2

    def get_rhs(self):
        return self.d
        
    def __call__(self, p, jac=False):
        self.pars.free = p
        v = -self.H.T @ self.pars.full + self.vals
        
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            return v, self.H[i,:].T
        return v


class ToySplineRegularization():
    
    def __init__(self, pars, mu=1.E-8):
        self.pars = pars
        self.mu = mu

        # build the regularization matrix
        n = len(self.pars['theta'].full)
        data = np.ones(n).astype(float)
        M = self.mu * scipy.sparse.dia_matrix((data, [0]), shape=(n,n)).tocoo()

        # build the large matrix which corresponds to the model parameters
        pp = pars.copy()
        pp.release()
        N = len(pp.full)
        i = pp['theta'].indexof(M.row)
        j = pp['theta'].indexof(M.col)
        idx = (i>=0) & (j>=0) # should be all true's
        self.M = scipy.sparse.coo_matrix((M.data[idx], (i[idx],j[idx])), shape=(N,N)).tocsr()
        
    def __call__(self, p, jac=True):
        self.pars.free = p
        v = float(self.pars.full.T @ self.M @ self.pars.full)
        if jac:
            i = np.where(self.pars.indexof() >= 0)[0]
            # there might be a more efficient way to slice the matrix ...
            return v, self.M[i,:][:,i]
        return v

    
def main(N=1000, minimize=minimize_nr_var_quad, mu_regularization=500., plot=False, seed=None):
    """
    """
    lc_data = salt3.generator.generate_lc(N, tmax_range=(0., 0.),
                                          sig_range=(0., 0.),
                                          norm_range=(1000., 1000.),
                                          seed=seed)
    
    model = ToySplineModel(lc_data, grid_size=250) # phase_range=[lc_data.date.min(), lc_data.date.max()]
    f = salt3.models.VarModelResiduals(lc_data, model)
    reg = ToySplineRegularization(f.pars, mu=mu_regularization)
    cons = ToySplineNonIntegralConstraints(model.pars, model.spline_basis)

    # rough initialization 
    f.pars['flux'].full[:] = 1000.
    f.pars['theta'].full[:] = 1.E-6
    
    f.pars['flux'].fix()
    x = minimize(f, f.pars.free, reg=reg, cons=cons)
    x = x[0]
    f.pars.free = x
    
    # release everybody and trash the fluxes a little 
    f.pars['flux'].release()
    f.pars['flux'].full[:] = 5000 + np.random.uniform(-1000, 1000, len(f.pars['flux'].full))
    x = minimize(f, f.pars.free, cons=cons, reg=reg, n_iter=1000)
    x = x[0]
    f.pars.free = x
    
    
    if plot:
        import pylab as pl
        import matplotlib as mpl
        
        pl.figure()
        gs = mpl.gridspec.GridSpec(4,1)
        ax = pl.subplot(gs[0:3,0])
        pl.errorbar(lc_data.date, lc_data.flux, yerr=lc_data.flux_err, ls='', marker='.', color='b')
        xx = np.linspace(lc_data.date.min(), lc_data.date.max(), 100)
        xx.sort()

        try:
            tmax = f.pars['tmax'].full
        except:
            tmax = np.zeros(N)
        xdata = salt3.plots.get_xdata_for_model_evaluation(N, tmax, lc_data.band_set, npoints=200)
        v = model(x, dp=xdata, jac=False)
        for sn in lc_data.sn_set:
            idx = xdata.sn == sn
            pl.plot(xdata.date[idx], v[idx], ls='-', color=pl.cm.jet(sn/xdata.sn.max()))
        
        pl.subplot(gs[3,0], sharex=ax)
        pl.errorbar(lc_data.date, f(x,jac=False)*lc_data.flux_err, yerr=lc_data.flux_err, marker='.', ls='', color='b')
        pl.grid(1)
        pl.xlabel('x')
        pl.ylabel('res')
        pl.subplots_adjust(hspace=0.05)

        pl.figure()
        xx = np.linspace(-50., 50., 10000)
        jacobian = model.spline_basis.eval(xx)
        pl.plot(xx, jacobian @ model.pars['theta'].full, 'r-')
        pl.xlabel('x')
        pl.ylabel('LC template (spline model only)')
        pl.grid(1)
        
    
    return lc_data, f, cons, reg, x


def test_model_derivatives(N=200, seed=0):
    """
    """
    lc_data = salt3.generator.generate_lc(N=N, tmax_range=(-70., 70.),
                                          sig_range=(-1., 1.),
                                          norm_range=(10., 1000.),
                                          seed=seed)

    model = ToySplineModel(lc_data, grid_size=250) # phase_range=[lc_data.date.min(), lc_data.date.max()], grid_size=100)
    f = salt3.models.ModelResiduals(lc_data, model)
    
    jacobian, JJ = check_grad(model, model.pars.free)
    D = np.abs(jacobian-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    J, JJ = check_grad(f, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model.pars.free[:] += np.random.uniform(-1., 1., len(model.pars.free))
    J, JJ = check_grad(model, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    J,JJ = check_grad(f, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)

    model.pars.full[:] += np.random.uniform(-1., 1., len(model.pars.full))
    J, JJ = check_grad(model, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)
    
    J,JJ = check_grad(f, model.pars.free)
    D = np.abs(J-JJ)
    np.testing.assert_allclose(np.zeros_like(D), D, atol=1.E-6)


# def test_toy1dsplinemodelfit_nr():
#     """
#     tests of the main (current) minimizer 
#     """
#     x_target = np.load(locate_test_datafile('test_toy1dsplinemodel_solution.npy'))
    
#     lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr_var_quad,
#                                mu_regularization=500., plot=0, seed=0)
#     np.testing.assert_allclose(x_target, x, atol=1.E-6)
#     np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)    
    

def test_toy1dsplinemodelfit_nr_qr():
    """
    tests of the older minimizer : minimizer_nr_qr (when there are
    constraints, this one does not scale very well with the size of
    the parameter vector. Indeed, the sparsity of the Q,R
    decomposition is not guaranteed, even with sparse constraints).
    """
    x_target = np.load(locate_test_datafile('test_toy1dsplinemodel_solution.npy'))
    lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr_qr, mu_regularization=500., plot=0, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)

    
def test_toy1dsplinemodelfit_nr_qr(datadir):
    """
    tests of the older minimizer : minimizer_nr_qr (when there are
    constraints, this one does not scale very well with the size of
    the parameter vector. Indeed, the sparsity of the Q,R
    decomposition is not guaranteed, even with sparse constraints).
    """
    x_target = np.load(locate_test_datafile('test_toy1dsplinemodel_solution.npy'))
    lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr_qr, mu_regularization=500., plot=0, seed=0)
    np.testing.assert_allclose(x_target, x, atol=1.E-6)
    np.testing.assert_allclose(np.zeros_like(cons(x, jac=False)), cons(x, jac=False), atol=1.E-8)


    
if __name__ == '__main__':
    lc, f, cons, reg, x = main(N=1000, minimize=minimize_nr_var_quad, mu_regularization=500., plot=1)



import pylab as pl

def compare_pars(tds, pars):
    """
    compares the parameters found in the sncosmo and after the NaCl fit
    """
    
    x0_tds = tds.sn_data.x0
    x1_tds = tds.sn_data.x1
    c_tds = tds.sn_data.col
    tmax_tds = tds.sn_data.tmax
    z = tds.sn_data.z
    
    x0_mod = pars['X0'].full[tds.sn_data.sn_index]
    x1_mod = pars['X1'].full[tds.sn_data.sn_index]
    c_mod = pars['col'].full[tds.sn_data.sn_index]
    tmax_mod = pars['tmax'].full[tds.sn_data.sn_index]
    
    pl.figure()
    pl.scatter(x0_tds, x0_mod)
    pl.title('X0 NaCl vs X0 sncosmo')
    pl.xlabel('X0 sncosmo')
    pl.ylabel('X0 NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(x1_tds, x1_mod)
    pl.title('X1 NaCl vs X1 sncosmo')
    pl.xlabel('X1 sncosmo')
    pl.ylabel('X1 NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(c_tds, c_mod)
    pl.title('col NaCl vs col sncosmo')
    pl.xlabel('col sncosmo')
    pl.ylabel('col NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(tmax_tds, tmax_mod)
    pl.title('tmax NaCl vs tmax sncosmo')
    pl.xlabel('tmax sncosmo')
    pl.ylabel('tmax NaCl')
    pl.show()
    
    pl.figure()
    pl.scatter(z, x0_mod/x0_tds, color='red')
    pl.title('X0 NaCl/X0 sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('X0 NaCl/X0 sncosmo')
    pl.show()

    pl.figure()
    pl.scatter(z, x1_mod/x1_tds, color='red')
    pl.title('X1 NaCl/X1 sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('X1 NaCl/X1 sncosmo')
    pl.show()
    
    pl.figure()
    pl.scatter(z, c_mod/c_tds, color='red')
    pl.title('col NaCl/col sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('col NaCl/col sncosmo')
    pl.show()
    
    pl.figure()
    pl.scatter(z, tmax_mod/tmax_tds, color='red')
    pl.title('tmax NaCl/tmax sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('tmax NaCl/tmax IDR')
    pl.show()

    pl.figure()
    pl.scatter(z, x1_mod-x1_tds, color='red')
    pl.title('X1 NaCl-X1 sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('X1 NaCl-X1 sncosmo')
    pl.show()
    
    pl.figure()
    pl.scatter(z, c_mod-c_tds, color='red')
    pl.title('col NaCl-col sncosmo vs z')
    pl.xlabel('z')
    pl.ylabel('col NaCl-col sncosmo')
    pl.show()
    
def compare_all_pars(old_pars, new_pars, block_name=['X0', 'X1', 'col', 'tmax', 'M0', 'M1', 'CL']):
    """
    """
    for block in block_name:
        fig, axes = pl.subplots(nrows=1, ncols=2, figsize=(10,10))
        fig.suptitle(f'parameter ' + block + ' before and after fit')
        axes[0].plot(old_pars[block].full, new_pars[block].full, 'k.')
        axes[0].set_xlabel(block + ' before fit')
        axes[0].set_ylabel(block + ' after fit')

        axes[1].plot(old_pars[block].full, (old_pars[block].full - new_pars[block].full)/old_pars[block].full, 'k.')
        axes[1].set_xlabel(block + ' before fit')
        axes[1].set_ylabel(block + ' before - after / before')



import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)

from scipy import sparse
import numpy as np
import pylab as pl
import pandas
from nacl.lib.fitparameters import FitParameters
from nacl.dataset import TrainingDataset
from nacl.models.salt import SALT2Like
from nacl.models.constraints import LinearConstraint, ConstraintSet
from nacl.models.regularizations import NaClSplineRegularization
from nacl.minimize import LogLikelihood2, Minimizer
import helpers



class VarianceModel:

    def __init__(self, model):
        self.training_dataset = model.training_dataset
        self.data = model.data

    def get_struct(self):
        return [('sigma', 1)]

    def init_pars(self, pars):
        self.pars = pars
        self.pars['sigma'].full[:] = 0.5

    def __call__(self, jac=False, model_flux=None, model_jac=None):
        """
        """
        N = len(self.data)
        sigma = self.pars['sigma'].full[0]
        v = np.full(N, sigma**2)
        if not jac:
            return v

        i = np.arange(N)
        j = np.full(N, self.pars['sigma'].indexof())
        vv = np.full(N, 2. * sigma)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((vv[idx], (i[idx], j[idx])),
                               shape=(N, n_free_pars))
        return v, J

    def noise(self, N):
        v = self(jac=False)
        return np.random.normal(scale=np.sqrt(v))


class AixBiCDTrainingDataset:
    """Simplistic training dataset for tests
    """
    def __init__(self, nlc=10, npts=100,
                 a_i=None, b_i=None,
                 sigma_meas=0.1, sigma_unknown=1.):
        """constructor - initialize a recarray with simulated data
        """
        self.nlc = nlc
        self.a_i = a_i
        if a_i is None:
            self.a_i = np.random.uniform(- 5.,  5., nlc)
        self.b_i = b_i
        if b_i is None:
            self.b_i = np.random.uniform(-10., 10., nlc)
        # self.c = - self.b_i.sum()
        # self.b_i -= self.b_i.sum()
        self.c = 0.
        self.npts = npts
        self.sigma_meas = sigma_meas
        self.sigma_unknown = sigma_unknown

        i_, x_, y_, ey_, valid_ = [], [], [], [], []
        for i in range(nlc):
            x = np.random.uniform(-1, 1, npts)
            sigma_tot = np.sqrt(sigma_meas**2 + sigma_unknown**2)
            noise = np.random.normal(scale=sigma_tot, size=npts)
            y = self.a_i[i] * x + self.b_i[i] + self.c + noise
            ey = np.full(npts, sigma_meas)
            i_.append(np.full(npts, i))
            x_.append(x)
            y_.append(y)
            ey_.append(ey)
            valid_.append(np.full(npts, 1))

        self.data = np.rec.fromarrays(
            (np.hstack(i_), np.hstack(x_), np.hstack(y_),
             np.hstack(ey_), np.hstack(valid_)),
                names=('i', 'x', 'y', 'ey', 'valid'))

    def nb_meas(self, valid_only=False):
        if valid_only:
            return self.data['valid'].sum()
        return len(self.data)

    def get_all_fluxes(self):
        return self.data['y']

    def get_all_fluxerr(self):
        return self.data['ey']

    def get_valid(self):
        return self.data['valid']

    def plot(self, ax=None, model=None):
        if ax is None:
            fig, axes = pl.subplots(nrows=1, ncols=1, figsize=(8,8))
        axes.errorbar(self.data['x'], self.data['y'],
                      yerr=self.data['ey'],
                      ls='', marker='.', color='k')
        if model is not None:
            a = model.pars['a'].full[0]
            b = model.pars['b'].full[0]
            pl.plot(self.data['x'], a*self.data['x']+b, 'r-')
        return axes


class AiBiCDModel:

    def __init__(self, training_dataset, var_class=None):
        self.training_dataset = training_dataset
        self.data = training_dataset.data
        self.variance_model = None
        if var_class is not None:
            self.variance_model = var_class(self)
        self.pars = self.init_pars()
        if self.variance_model is not None:
            self.variance_model.init_pars(self.pars)

    def get_struct(self):
        nlc = self.training_dataset.nlc
        return [('a_i', nlc), ('b_i', nlc), ('c', 1)]  # ('d', 1)]

    def init_pars(self):
        s = self.get_struct()
        if self.variance_model:
            s += self.variance_model.get_struct()
        fp = FitParameters(s)
        fp['a_i'] = 1.
        return fp

    def __call__(self, p, jac=False):
        tds = self.training_dataset
        a_i = self.pars['a_i'].full[tds.data.i]
        b_i = self.pars['b_i'].full[tds.data.i]
        c = self.pars['c'].full[0]
        # d = self.pars['d'].full[0]
        v = a_i * self.data['x'] + b_i + c   # + d
        if not jac:
            return v

        model_vals = v

        N = len(self.data)
        i,j,v = [], [], []
        i.append(np.arange(N))
        j.append(np.full(N, self.pars['a_i'].indexof(tds.data.i)))
        v.append(self.data['x'])

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['b_i'].indexof(tds.data.i)))
        v.append(np.full(N, 1.))

        i.append(np.arange(N))
        j.append(np.full(N, self.pars['c'].indexof()))
        v.append(np.full(N, 1))

        # i.append(np.arange(N))
        # j.append(np.full(N, self.pars['d'].indexof()))
        # v.append(np.full(N, 1))

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        idx = j >= 0
        n_free_pars = len(self.pars.free)
        J = sparse.coo_matrix((v[idx], (i[idx],j[idx])),
                              shape=(N,n_free_pars))

        return model_vals, J



# class ModelBCD2Constraint(Constraint):

#     def __init__(self, model, rhs, pars_struct=None):
#         super().__init__(model, rhs, pars_struct)

#     def __call__(self, p=None, deriv=False):
#         """
#         """
#         pars = self.model.pars
#         cons = pars['b_i'].full**2 + pars['c'].full**2 + pars['d'].full**2
#         cons = cons.sum()
#         cons -= self.rhs

#         if not deriv:
#             return cons

#         # if deriv
#         npars_full = len(self.model.pars.full)
#         j = np.hstack((self.pars_struct['b_i'].indexof(),
#                        self.pars_struct['c'].indexof(),
#                        self.pars_struct['d'].indexof()))
#         v = np.hstack((2. * pars['b_i'].full,
#                        2. * pars['c'].full,
#                        2. * pars['d'].full))
#         i = np.zeros(len(j))
#         dcons = sparse.coo_matrix((v, (i,j)), shape=(1, npars_full))

#         # second derivatives
#         d2cons = sparse.coo_matrix((np.full(len(j), 2.), (j,j)),
#                           shape=(npars_full, npars_full))

#         return cons, dcons, d2cons



def sum_bi(model, pars_struct=None):
    """
    """
    pars = model.pars
    npars = len(pars.full)

    i, j, v = [], [], []
    j = pars['b_i'].indexof()
    i = np.full(len(j), 0)
    v = np.full(len(j), 1.)
    mat = sparse.coo_matrix((v, (i,j)), shape=(1, npars)).tocsr()

    return LinearConstraint(model, mat, 0., pars_struct)


# def test_aibicd_d_fixed_linear_constraint():
#     tds = AixBiCDTrainingDataset(sigma_unknown=2.1)
#     model = AiBiCDModel(tds, var_class=VarianceModel)
#     model.pars['d'].fix()
#     c = sum_bi(model)
#     cons = ConstraintSet(model, [c])
#     ll = LogLikelihood2(model, variance_model=model.variance_model, cons=[cons])
#     minz = Minimizer(ll)
#     ret = minz.minimize(model.pars.free)
#     ret['model'] = model
#     ret['cons'] = cons
#     return ret


def plot_reco(rec_data, tds, par_num=0, title=''):

    fig, axes = pl.subplots(figsize=(8,8), nrows=2, ncols=3)

    delta = rec_data.a_i.apply(lambda x: x[par_num]) - tds.a_i[par_num]
    axes[0,0].hist(delta, bins=20)
    axes[0,0].axvline(ls=':', color='r')
    axes[0,0].set_xlabel('$\Delta a$')

    delta = rec_data.b_i.apply(lambda x: x[par_num]) - tds.b_i[par_num]
    axes[0,1].hist(delta, bins=20)
    axes[0,1].axvline(ls=':', color='r')
    axes[0,1].set_xlabel('$\Delta b$')

    delta = rec_data.c - tds.c
    axes[0,2].hist(delta, bins=20)
    axes[0,2].axvline(ls=':', color='r')
    axes[0,2].set_xlabel('$\Delta c$')

    delta = rec_data.sigma - tds.sigma_unknown
    axes[1,0].hist(delta, bins=20)
    axes[1,0].axvline(ls=':', color='r')
    axes[1,0].set_xlabel('$\Delta \sigma$')

    axes[1,1].hist(rec_data.cons, bins=20)
    axes[1,1].set_xlabel('constraints')

    axes[1,2].hist(rec_data.chi2/rec_data.ndof)
    axes[1,2].set_xlabel('$\chi^2 / ndof$')

    if title:
        fig.suptitle(title)


def main_aixbicd(nlc=10, n_realizations=1, npts=100, plot=False,
                 a_i=None, b_i=None,
                 sigma_meas=0.25, sigma_unknown=1.2):
    """
    """
    ret = []
    if a_i is None:
        a_i = np.random.uniform(-10., 10., nlc)
    if b_i is None:
        b_i = np.random.uniform(-10., 10., nlc)
        b_i -= b_i.mean()
    for i in range(n_realizations):
        tds = AixBiCDTrainingDataset(nlc=nlc, npts=npts,
                                     a_i=a_i, b_i=b_i,
                                     sigma_meas=sigma_meas,
                                     sigma_unknown=sigma_unknown)
        model = AiBiCDModel(tds, var_class=VarianceModel)
        cons = ConstraintSet(model, [sum_bi(model)])
        ll = LogLikelihood2(model,
                            variance_model=model.variance_model,
                            cons=[cons])
        minz = Minimizer(ll)
        # n_free_pars = len(model.pars.free)
        n_pars = len(model.pars.full)
        model.pars.full[:] = np.random.uniform(-100., 100, n_pars)
        model.pars['sigma'].full[:] = 10.
        res = minz.minimize(model.pars.free)
        ret.append(((np.array(list(map(float, model.pars['a_i'].full))),
                   np.array(list(map(float, model.pars['b_i'].full))),
                   np.array([model.pars['c'].full[0]]),
                   np.array([model.pars['sigma'].full[0]]),
                   float(cons(model.pars.free)),
                   float(res['chi2']),
                   float(res['ndof']),
                   res['status'])))

    cols = ['a_i', 'b_i', 'c', 'sigma']
    cols += ['cons', 'chi2', 'ndof', 'status']
    ret = pandas.DataFrame.from_records(ret, columns=cols,
                                        coerce_float=True)

    if plot:
        plot_reco(ret, tds,
                  title=f'AixBiCDModel [{n_realizations} realizations]')

    return ret, tds


if __name__ == '__main__':
    pass




import numpy as np
import pylab as pl
pl.ion()
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)
from nacl.dataset import TrainingDataset
from nacl.train import TrainSALT2Like
from nacl.models import salt2
import pandas as pd
import os
os.environ['KMP_DUPLICATE_LIB_OK']='True'
from lemaitre import bandpasses
import release

#will be removed from the full fit
from nacl.specutils import Spec, clean_and_project_spectra

if __name__ == '__main__':
    fl = bandpasses.get_filterlib()  #Initialise filters
    tds = release.dataset_release('lemaitre_light') # Uncompressed dataset
    model_ = salt2.get_model(tds)
    
    #will be removed later : Compression should be done beforehand
    #I am leaving it here because the compression takes 2 seconds here
    logging.info('compression...')
    projected_spectra, in_error = clean_and_project_spectra(tds, model_.basis.bx)
    logging.info('done: in_error: {len(in_error)}')

    ptds = TrainingDataset(tds.sn_data.nt, lc_data=tds.lc_data.nt,
                           spec_data=np.rec.array(np.hstack(projected_spectra)),
                           basis=model_.basis.bx,
                           filterlib=fl)
    
    #TRAINING
    trainer = TrainSALT2Like(ptds, variance_model='simple_snake')  #Initialise with an error model
    #trainer = TrainSALT2Like(ptds, variance_model='local_snake')
    #trainer = TrainSALT2Like(ptds, variance_model='sn_lambda_snake')
    #trainer = TrainSALT2Like(ptds, variance_model='sn_local_snake')
    trainer.train_salt2_model(save=True, path='test_mini')  #train
    
    
    #RESULTS
    pars_trained = trainer.log[-1].pars  #parameters
    v_trained = trainer.log[-1].v  #model prediction
    #cov_matrix = trainer.log[-1].minz.get_cov_matrix()[0] #full covariance matrix (this can take a few minutes)
    
    #PLOTS
    trainer.plot_lc('ZTF18aapsedq', numfit=3, plot_variance=True)



import os
import os.path as op
import numpy as np

from nacl.models.salt import SALT2Like
from nacl.models import variancemodels
import nacl.models as models
from nacl.simulations.fullsim import cadences, sngen, snsurvey
from nacl.simulations import etc
# from nacl.simulations.flux_simulator import FluxSimulator
from nacl.simulations import FluxSimulator, SpecUncertainties


def locate_test_datafile(fn, dirname='expected', check=False):
    """
    """
    ret = op.join(op.dirname(__file__), dirname, fn)
    if check and not op.isfile(ret):
        raise(FileNotFoundError(ret))
    return ret


def generate_dataset(nsn=10, seed=42, string_ids=True,
                     calib_variance=0., compress=False):
    """generate a model and a training dataset
    """
    # first, we instantiate s SN generator
    # it generates the SN distribution, according to the chosen cosmology
    gen = sngen.DVDzSNGenerator(string_ids=string_ids,
                                alpha=-0.13, beta=3., sigma_int=0.1)

    # using this generator, we generate a supernova sample
    sample = gen.generate_sample(nsn=nsn)

    # then, we need to observe these supernovae,
    # to generate follow-up data. We generate the
    # main survey cadence...
    cad = cadences.MetronomicCadenceFactory(cadences.lsst_ddf_ideal,
                                            n_fields=1)
    obslog = cad()

    # then, we instantiate the survey class.
    # i.e. something that,
    survey = snsurvey.FieldBasedSNSurvey(sample, obslog,
        visits=[snsurvey.StandardSpectroscopicObservations()])

    # and generate the training dataset
    # at this stage, we have not generated the fluxes yet.
    training_dataset = survey.observe()
    
    # mag sky values are a little too optimistic
    # TODO: move this hack somewhere else
    training_dataset.lc_data.nt['mag_sky'] -= 1.5
    training_dataset.lc_data.nt['seeing'] = 1.

    # now, instantiate a model
    # model = SALT2Like(training_dataset,
    #                   init_from_salt2_file='salt2.npz',
    #                   normalization_band_name='SWOPE::B',
    #                   error_snake_model_type=variancemodels.SimpleErrorSnake)
    #                   # calib_variance=calib_variance)
    model = models.get_model(training_dataset)
    model.init_pars()
    model.init_from_training_dataset()
    
    # generate an instance of the training dataset
    fs = FluxSimulator(model)
    fs.model.pars['sigma_snake'].full[:] = 0.05
    fs.update(p=model.pars.free, 
              phot_etc=etc.find('LSST'),
              spec_etc=SpecUncertainties(),
              phot_uncertainty_pedestal=None) # was 0.01
    tds = fs()

    # and instantiate a model around it
    tds_model = SALT2Like(tds,
                          init_from_salt2_file='salt2.npz',
                          normalization_band_name='SWOPE::B',
                          error_snake_model_type=variancemodels.SimpleErrorSnake)
    tds_model.init_from_training_dataset()
    tds_model.pars['sigma_snake'].full[:] = 0.05

    # and simulate the fluxes
    # fs = FluxSimulator(model, seed=seed, salt2_filename='salt2.npz',
    #                    instrument_model=etc.find('LSST'),
    #                    calib_scatter=model.calib_scatter,
    #                    color_scatter=model.color_scatter,
    #                    error_snake=model.error_snake)
    # fs.load_truth_from_tds()
    # fs.update_fluxes()
    # fs.update_photometric_uncertainties()
    # fs.update_spectroscopic_uncertainties()
    # fs.add_measurement_noise()

    # remove failed measurements ?
    if compress:
        training_dataset.compress()
        model = SALT2Like(training_dataset,
                          init_from_salt2_file='salt2.npz',
                          normalization_band_name='SWOPE::B',
                          error_snake_model_type=variancemodels.SimpleErrorSnake)
        model.init_pars()

    return tds, tds_model


def check_grad(model, p):
    """
    """
    v,jacobian = model(p, jac=True)
    dx = 1.E-7
    p0 = p.copy()
    df = []
    for i in range(len(p)):
        p[i] += dx
        vp = model(p, jac=False)
        df.append((vp-v)/dx)
        p[i] -= dx
    return np.array(jacobian.todense()), np.vstack(df).T


def check_deriv(pen, p):
    """
    """
    v, grad, hess = pen(p=p, deriv=True)
    dx = 1.E-5
    p0 = p.copy()

    df =[]
    for i in range(len(p0)):
        p0[i] += dx
        # vp = pen(p0, deriv=False)-
        p0[i] -= (2*dx)
        vm = pen(p0, deriv=False)
        df.append((vp-vm)/(2*dx))
        d2f.append((vp - 2*v + vm)/dx**2)
        p0[i] += dx
    return np.array(grad), np.vstack(df).T


def check_grad_var(func, p, g):
    """
    """
    h = 1.e-7
    val0, J0 = func.model(p, jac = True)
    v0, V0 = func.VarianceModel(g, p, jacobian = J0)#val0,

    df = []
    dmod = []
    for i in range(len(p)+len(g)):
        #print(i)
        if i < len(p):
            p[i] += h
        else :
            g[i-len(p)] += h

        val1 = func.model(p)
        v1 = func.VarianceModel(g, p)#val1,
        #print(i)
        dv = (v1-v0)/h
        #print(V0.tocsr()[i].toarray().squeeze(), dv)
        #np.testing.assert_allclose(V0.tocsr()[i].toarray().squeeze(), dv, rtol=1.E-2)
        df.append(dv)
        if i < len(p):
            dmod.append((val1-val0)/h)
        if i < len(p):
            p[i] -= h
        else :
            g[i-len(p)] -= h

    return V0.todense(), np.vstack(df).T, J0.todense(), np.vstack(dmod).T



#!/usr/bin/env python3


import numpy as np
import numpy.polynomial.polynomial as nppol
from scipy import sparse


class CalibPrior:

    def __init__(self, calib_covmat, block_name='eta'):
        self.calib_covmat = calib_covmat
        self.matrix = None
        self.block_name = block_name

    def init_pars(self, pars):
        self.matrix = self.init_w_matrix(pars)

    def init_w_matrix(self, pars):
        """
        """
        W = np.linalg.inv(self.calib_covmat)
        W = sparse.coo_matrix(W)
        pars_full = pars.copy()
        pars_full.release()

        N = len(pars_full.full)
        i = pars[self.block_name].indexof(W.row)
        j = pars[self.block_name].indexof(W.col)
        M = sparse.coo_matrix((W.data, (i,j)), shape=(N,N)).tocsr()

        return M

    def __call__(self, pars, deriv=False):
        """
        """
        if self.matrix is None:
            self.matrix = self.init_w_matrix(pars)
        pp = pars.full
        penalty = float(pp.T @ self.matrix @ pp)

        if not deriv:
            return penalty

        n = self.matrix.shape[0]
        idx = pars.indexof(np.arange(n)) >= 0

        grad = 2. * (self.matrix @ pp)[idx]
        hess = 2. * self.matrix[:,idx][idx,:]

        return penalty, grad, hess


class ColorScatterPrior:

    def __init__(self, model, pivot_wl=5000., deg=2):
        """
        """
        self.model = model
        self.pivot_wl = pivot_wl
        self.deg = deg
        self.mean_sn_band_wl = None
        # a piece of code to compute the
        # equivalent wavelength associated to each kappa
        dp = model.data
        self.kappa_wl = np.unique(np.vstack((dp.sn_band_index, dp.wl)).T, axis=0)[:,1]

    def get_struct(self):
        return [('sigma_kappa', self.deg+1)]

    def init_pars(self, pars):
        """init the sigma-kappa parameters
        """
        pars['sigma_kappa'].full[:] = 1.E-6
        pars['sigma_kappa'].full[0] = -4.

    def __call__(self, pars, deriv=False):
        """evaluate the color scatter penalty (along with its derivatives)
        """
        sigma_kappa = pars['sigma_kappa'].full
        kappa = pars['kappa'].full

        # variance model
        rwl = (self.kappa_wl - self.pivot_wl) / self.pivot_wl
        var = np.exp(nppol.Polynomial(sigma_kappa)(rwl))

        # log det, penaly
        log_det = np.log(var).sum()
        quad_penalty = (kappa**2 / var).sum()
        penalty = log_det + quad_penalty

        if not deriv:
            return penalty

        # gradient w.r.t kappa
        N = len(pars.free)
        grad = np.zeros(N)
        j = pars['kappa'].indexof()
        idx = j>=0
        grad[j[idx]] = (2. * kappa / var)[idx]

        # gradient w.r.t sigma_kappa
        nk = len(pars['kappa'].full)
        ns = len(pars['sigma_kappa'].full)
        var = var.reshape(-1,1)
        kappa = kappa.reshape(-1,1)
        dq_ds = nppol.polyvander(rwl, deg=self.deg)
        tr_wdv =  np.sum(dq_ds, axis=0)
        kwdvwk = -np.sum(kappa**2 * dq_ds / var, axis=0)
        print(tr_wdv, kwdvwk)
        j = pars['sigma_kappa'].indexof()
        idx = j >= 0
        # check the signs
        grad[j[idx]] = (-tr_wdv - kwdvwk)[idx]

        # hessian w.r.t kappa
        i, j, val = [], [], []
        m = sparse.coo_matrix(np.diag(2. / var.flatten()), shape=(nk,nk))
        i.append(pars['kappa'].indexof(m.row))
        j.append(pars['kappa'].indexof(m.col))
        val.append(m.data)

        # hessian: kappa-sigma_kappa block
        m = sparse.coo_matrix(-2. * kappa * dq_ds / var, shape=(nk,ns)) # was -2
        i.append(pars['kappa'].indexof(m.row))
        j.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)
        j.append(pars['kappa'].indexof(m.row))
        i.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)

        # hessian: sigma_kappa - sigma_kappa block
        dq = kappa * dq_ds / np.sqrt(var)
        # tr_wdvwdv = -4 * dq.T @ dq
        # dq = kappa**2 * dq_ds / sig
        kwdvwdvwk = dq.T @ dq
        m = sparse.coo_matrix(kwdvwdvwk, shape=(ns,ns))
        i.append(pars['sigma_kappa'].indexof(m.row))
        j.append(pars['sigma_kappa'].indexof(m.col))
        val.append(m.data)

        # kk = dsig_ds / sig
        # tr_wdvwdv = -4 * kk.T @ kk
        # kk = kappa * dsig_ds / sig**2
        # kwdvwdvwk = 6 * kk.T @ kk

        # m = sparse.coo_matrix(tr_wdvwdv + kwdvwdvwk)
        # i.append(pars['sigma_kappa'].indexof(m.row))
        # j.append(pars['sigma_kappa'].indexof(m.col))
        # val.append(m.data)

        # encapsulate all that into the (much larger) parameter vector space
        i = np.hstack(i)
        j = np.hstack(j)
        val = np.hstack(val)
        idx = (i>=0) & (j>=0)
        n_free_pars = len(pars.free)
        hessian = sparse.coo_matrix((val[idx], (i[idx], j[idx])), shape=(n_free_pars, n_free_pars))

        return penalty, grad, hessian



"""Constraints
"""

import numpy as np
from scipy import sparse
from ...constraints import LinearConstraint, ConstraintSet



class MaxLcAtPhase(LinearConstraint):

    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        J = self.model.basis.eval(np.array([self.phase]), np.array([self.phase])).tocoo()
        pp = pars.copy()
        pp.release()
        N = len(J.col)
        i = np.full(N, 0)
        j = pp['theta'].indexof(J.col)
        # idx = j>=0
        npars = len(pp.full)
        J = sparse.coo_matrix((J.data, (i, j)), shape=(1, npars))
        return J

    # def max_lc_at_phase(model, rhs, phase=0.):
    #     """
    #     """
    #     J =  model.basis.eval(np.array([phase])).toarray()
    #     return LinearConstraint(J, rhs)

class DMaxLcAtPhase1(LinearConstraint):

    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        J, _ = self.model.basis.gradient(np.array([self.phase]), np.array([self.phase]))
        J = J.tocoo()
        pp = pars.copy()
        pp.release()
        N = len(J.col)
        i = np.full(N, 0)
        j = pp['theta'].indexof(J.col)
        npars = len(pp.full)
        J = sparse.coo_matrix((J.data, (i, j)), shape=(1, npars))
        return J

class DMaxLcAtPhase2(LinearConstraint):

    def __init__(self, model, rhs, phase):
        super().__init__(model, rhs)
        self.phase = phase

    def init_h_matrix(self, pars):
        _, J = self.model.basis.gradient(np.array([self.phase]), np.array([self.phase]))
        J = J.tocoo()
        pp = pars.copy()
        pp.release()
        N = len(J.col)
        i = np.full(N, 0)
        j = pp['theta'].indexof(J.col)
        npars = len(pp.full)
        J = sparse.coo_matrix((J.data, (i, j)), shape=(1, npars))
        return J

class MeanColor(LinearConstraint):

    def __init__(self, model, rhs):
        super().__init__(model, rhs)

    def init_h_matrix(self, pars):
        nsn = len(pars['x0'].full)
        npars = len(pars.full)
        j = pars['color'].indexof(np.arange(nsn))
        i = np.full(nsn, 0)
        v = np.full(nsn, 1./nsn)
        idx = j >= 0
        J = sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(1, npars))

        return J

def cons(model, mu=1.E6, color=False):
    """
    """
    #    m0 = max_lc_at_phase(model, 0., phase=0.)
    #    dm0 = dmax_lc_at_phase(model, 0., phase=0.)
    m0 = MaxLcAtPhase(model, 1., 0.)
    dm0_1 = DMaxLcAtPhase1(model, 0., 0.)
    dm0_2 = DMaxLcAtPhase2(model, 0., 0.)
    if not color:
        return ConstraintSet([m0, dm0_1, dm0_2], mu=mu)
    else:
        c0 = MeanColor(model, 0.)
        return ConstraintSet([m0, dm0_1, dm0_2, c0], mu=mu)



import numpy as np
import scipy
import matplotlib.pyplot as pl
pl.ion()
from astropy.io import fits as pf
from scipy.sparse import coo_matrix, dia_matrix, dok_matrix, diags
from scipy.optimize import leastsq, fmin_ncg, check_grad, approx_fprime
from saunerie.lsqfit import Chi2
from saunerie.fitparameters import FitParameters
from saunerie.bspline import BSpline, CardinalBSplineC, integ
from croaks import NTuple
from numpy.polynomial.legendre import leggauss
from sksparse.cholmod import cholesky, cholesky_AAt
from saunerie import optim, plottools
from scipy.integrate import quad
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.DEBUG)
from astropy.modeling import models, fitting
from scipy.integrate import simps

# The model y is a linear function of x.
# In all this file, we define differents models for a variance. The variance may either be define as a polynomial function of x or as a linear function of f(x), f is a linear function of x.
# There are four different type of variance : 3 polynomial function each labeled by the degre of the polynome n_gamma, it can be 1 (constant) , 2 (linear) or 4.
# The fourth one, is where the variance depends linearly of f(x), in the following it will be labelled by n_gamma = 'dep'  


# coefficient are define as follow :
# y = 1.235 * x + 9.123

# variance models :
  # polynomial function of x:
     # n_gamma = 2 :  var = 1.356
     # n_gamma = 2 :  var = 1.356 * x + 2.222
     # n_gamma = 4 :  var = 1/30. * (x**3 - 19. * x**2 + 133. * x + 229.)
  # linear function of f(x) :
     # n_gamma = 'dep' : var = 0.01 * f(x) + 0.005 

def variance(x, n):
    """
    create the "n" variance describes by the code just above.
    
    Input : - x = Array, argument of the variance function, it may be x (if n == &, 2 or 4) or y (if n == 'dep').
            - n = 1, 2, 4 or 'dep', code of the variance we want to create.
    
    Output : Array, the variance wanted.
    """
    if n == 1:
        return 1.356 * np.ones(x.shape)
    elif n == 2:
        return 1.325*x + 2.2222
    elif n == 4:
        return 1/30.*(x**3 - 19.*x**2 + 133.*x + 229.)
    elif n == 'dep':
        return 0.01 * x + 0.005
    
    
def gen_var(N, ng, a = 1.235, b = 9.123):
    """
    return the definition interval of the function, the linear function and the variance.
    
    Input : - N = int, number of points of the definition interval.
            - ng = 1, 2, 4, 'dep', the n_gamma.
            - a = float, the slope of the function.
            - b = float, the intercept.
    
    Output : - x = array(N), the definition interval.
             - y = array(N), linear function.
             - var = array(N), variance at each point.
    """
    x = np.linspace(1,10,N)
    mod = a * x + b
    if ng == 'dep':
        var = variance(mod, ng)
    else :
        var = variance(x, ng)
    n = []
    for i in range(N):
        n.append(np.random.normal(0, np.sqrt(np.abs(var[i])), 1)[0])
    n = np.array(n)
    y = a * x + b + n
    return x, y, var



class SModel(object):
    """
    definition of a new classe, a linear function : y = \alpha * x + \beta
    parameters are alpha and beta.
    """
    def __init__(self, x):
        self.nb_sn = 1 
        self.xdata = x
        self.pars = self.para_init()

    def para_init(self):
        fp = [("alpha", self.nb_sn), ("beta", self.nb_sn)]
        fp = FitParameters(fp)
        fp['alpha'] = 1.#1.235 
        fp['beta'] = 10. #9.123 
        return fp


    def __call__(self, p, jac=False, xdata=None):
        if xdata is None:
            xdata = self.xdata
        self.pars.free = p
        v = p[0] * xdata + p[1]
        if jac:
            N = len(xdata)

            # derivatives wrt alpha:
            ii_a = np.zeros((N))
            jj_a = np.arange(0,N)
            val_a = xdata

            # derivatives wrt Beta:
            ii_b = np.ones((N))
            jj_b = np.arange(0,N)
            val_b = np.ones((len(xdata)))

            i = np.hstack((ii_a, ii_b))
            j = np.hstack((jj_a, jj_b))
            val_jac = np.hstack((val_a, val_b))
            idx_fix = np.where(j != -1)
            Jac = coo_matrix((val_jac[idx_fix], (i[idx_fix],j[idx_fix])), shape = (len(self.pars.free),N))
            return  v , Jac.T
        else :
            return v

# This file has been written in a way such that the two variance model, the one depending on x and the one depening on f(x), are separated.
# They have their own class model and their own minimization fuction.
# To sinthetize :
   # polynomial function of x : - model : WModel
   #                            - minimization function : fit_sqp_lin
   #
   # linear function of f(x) : - model : WdepModel
   #                           - minimization function : fit_lin_var
# They are not the same, model return different element, one is the inverse of the other, and functions do not take exactly same inputs.

        
class WModel(object):
    """
    Model for a polynomial dependance of x for the variance.
    """
    def __init__(self, x, n):
        self.xdata = x
        self.n_g = n                                                                                                                                                                           
        self.pars = self.para_init()

    def para_init(self):
        if self.n_g == 4:
            fp = [("gamma3", 1), ("gamma2", 1), ("gamma1", 1), ("gamma0", 1)]
            fp = FitParameters(fp)
            fp['gamma3'] = 1./35.
            fp['gamma2'] = -0.63
            fp['gamma1'] = 4.43
            fp['gamma0'] = 7.63
            return fp

        elif self.n_g == 1:
            fp = [("gamma0", 1)]
            fp = FitParameters(fp)
            fp['gamma0'] = 1.5
            return fp
        
        elif self.n_g == 2:
            fp = [("gamma1", 1), ("gamma0", 1)]
            fp = FitParameters(fp)
            fp['gamma1'] = 1.5
            fp['gamma0'] = 2.
            2
            return fp

    def __call__(self, p):
        """
        return the matrix of the inverse of the variance on the diagonal.
        
        Inputs : - p = Array, parameters of the model.
        
        Output : - Array, 1/V equivalent of the matrix w.
        """
        if len(p) == 4:
            return np.diag(1/(p[3] + p[2] * self.xdata + p[1] * self.xdata**2 + p[0] * self.xdata**3))
        elif len(p) == 1:
            return np.diag(1/(p[0]*np.ones(self.xdata.shape)))
        elif len(p) == 2:
            return np.diag(1/(p[1] + p[0] * self.xdata))

# minimization function of the polynomial dependance

def fit_sqp_lin(y, x, variance, n_gamma = 2, max_iter = 100, ll = 0.1, dchi2_min = 0.1, reml = False, plot = True):
    """
    Minimization of the chi2.

    Inputs : - y = Array, the f(x) values.
            - x = Array, the definition interval.
            - variance = Array, the variance considered.
            - n_gamma = 1, 2, 4, degree of the polynome
            - max_iter = int, is the maximum number of iteration wanted.
            - ll = float, help to inverse the hessian matrix, by adding somthing one its diagonal.
            - dchi2_min = float, variation of chi2 desired to reach.
            - reml = bool, to consider the reml technic, NOT use here !
            - plot = bool, True will plot.

    Outputs : - p = Array, the final parameter of the linear model.
             - g  = Array, the final parameter of the variance model.
             - model = SModel, model of the straight line.  
             - var_model = WModel, model of the variance.
    """
    model = SModel(x)
    var_model = WModel(x, n = n_gamma)
    
    p = model.pars.free
    g = var_model.pars.free
    print(g)
    dp = np.zeros(p.shape)
    dg = np.zeros(g.shape)

    v, J = model(p, jac = True)
    w = var_model(g)

    res = y - v 
    
    chi2 = (np.linalg.multi_dot([res.T,w,res]).sum() )/ (len(y) - (p.shape[0] + g.shape[0]))#+ np.log(np.linalg.det(w))

    Dchi = 0.
    niter = 0
    
    while niter < max_iter:
        v, J = model(p, jac = True)
        w = var_model(g)

        res = y - v
        H = 2 * np.linalg.multi_dot([J.toarray().T, w, J.toarray()]) + ll * diags(np.ones(J.shape[1]))

        def construction(H = H, res = res, J = J, n_gamma = n_gamma, g = g ,reml = reml):
            """    
            Create Hessian matrix and the gradient vector needed to minimized the chi2, in order to fit models of the straight line and variances on data, using Newton-Raphson's method.
            """
            H = coo_matrix(H)
            H00 = H.data
            i00 = H.row
            j00 = H.col

            if n_gamma == 1 :
                H11 = (w.dot(w) * (2*g)**2).trace()
                i11 = p.shape[0] + n_gamma - 1
                j11 = p.shape[0] + n_gamma - 1

                val = np.hstack((H00, H11))
                i = np.hstack((i00, i11)) 
                j = np.hstack((j00, j11))
                H_l = coo_matrix((val, (i,j)), shape = (p.shape[0] + n_gamma, p.shape[0] + n_gamma))
            
                resul = np.hstack((2 * np.linalg.multi_dot([J.toarray().T, w,res]), 2*g*(np.linalg.multi_dot([res.T, w**2, res]) - w.trace())))
                return H_l, resul


            elif n_gamma == 2:
                X = np.diag(x)

                dev1 = 2 * (g[0] * X**2 + g[1] * X)
                dev0 = 2 * (g[0] * X + g[1]) 
                
                H11 = (np.linalg.multi_dot([w, dev1, w, dev1]).trace()) 
                i11 = [2]
                j11 = [2]
                
                H12 = np.linalg.multi_dot([w, dev1, w, dev0]).trace()
                i12 = [2]
                j12 = [3]

                H21 = np.linalg.multi_dot([w, dev0, w, dev1]).trace()
                i21 = [3]
                j21 = [2]
                
                H22 = np.linalg.multi_dot([w, dev0, w, dev0]).trace()
                i22 = [3]
                j22 = [3]


                val = np.hstack((H00, H11, H12, H21, H22))
                i = np.hstack((i00, i11, i12, i21, i22)) 
                j = np.hstack((j00, j11, j12, j21, j22)) 
                H_l = coo_matrix((val, (i,j)), shape = (p.shape[0] + n_gamma, p.shape[0] + n_gamma))

                resul = np.hstack((2 * np.linalg.multi_dot([J.toarray().T, w, res]), np.linalg.multi_dot([res.T, w, dev1, w, res]) - (w).dot(dev1).trace(), np.linalg.multi_dot([res.T, w, dev0, w, res]) - (w).dot(dev0).trace()))
                return H_l, resul
                    
            elif n_gamma == 4:
                X = np.diag(x)

                dev3 = 2 * ( g[0] * X**6 + g[1] * X**5 +  g[2] * X**4 + g[3] * X**3)
                dev2 = 2 * ( g[0] * X**5 + g[1] * X**4 +  g[2] * X**3 + g[3] * X**2)
                dev1 = 2 * ( g[0] * X**4 + g[1] * X**3 +  g[2] * X**2 + g[3] * X)
                dev0 = 2 * ( g[0] * X**3 + g[1] * X**2 +  g[2] * X + g[3])
                
                H11 = (np.linalg.multi_dot([w, dev3, w, dev3]).trace())
                i11 = [2]
                j11 = [2]
                
                H12 = np.linalg.multi_dot([w, dev3, w, dev2]).trace()
                i12 = [2]
                j12 = [3]

                H13 = np.linalg.multi_dot([w, dev3, w, dev1]).trace()
                i13 = [2]
                j13 = [4]

                H14 = np.linalg.multi_dot([w, dev3, w, dev0]).trace()
                i14 = [2]
                j14 = [4]
                
                H21 = np.linalg.multi_dot([w, dev2, w, dev3]).trace()
                i21 = [3]
                j21 = [2]

                H22 = np.linalg.multi_dot([w, dev2, w, dev2]).trace()
                i22 = [3]
                j22 = [3]

                H23 = np.linalg.multi_dot([w, dev2, w, dev1]).trace()
                i23 = [3]
                j23 = [4]

                H24 = np.linalg.multi_dot([w, dev2, w, dev0]).trace()
                i24 = [3]
                j24 = [5]

                H31 = np.linalg.multi_dot([w, dev1, w, dev3]).trace()
                i31 = [4]
                j31 = [2]

                H32 = np.linalg.multi_dot([w, dev1, w, dev2]).trace()
                i32 = [4]
                j32 = [3]
		
                H33 = np.linalg.multi_dot([w, dev1, w, dev1]).trace()
                i33 = [4]
                j33 = [4]

                H34 = np.linalg.multi_dot([w, dev1, w, dev0]).trace()
                i34 = [4]
                j34 = [5]

                H41 = np.linalg.multi_dot([w, dev0, w, dev3]).trace()
                i41 = [5]
                j41 = [2]
                
                H42 = np.linalg.multi_dot([w, dev0, w, dev2]).trace()
                i42 = [5]
                j42 = [3]
		
                H43 = np.linalg.multi_dot([w, dev0, w, dev1]).trace()
                i43 = [5]
                j43 = [4]

                H44 = np.linalg.multi_dot([w, dev0, w, dev0]).trace()
                i44 = [5]
                j44 = [5]
                

                val = np.hstack((H00, H11, H12, H13, H14, H21, H22, H23, H24, H31, H32, H33, H34, H41, H42, H43, H44))
                i = np.hstack((i00, i11, i12, i13, i14, i21, i22, i23, i24, i31, i32, i33, i34, i41, i42, i43, i44))
                j = np.hstack((j00, j11, j12, j13, j14, j21, j22, j23, j24, j31, j32, j33, j34, j41, j42, j43, j44))
                H_l = coo_matrix((val, (i,j)), shape = (p.shape[0] + n_gamma, p.shape[0] + n_gamma))

                resul = np.hstack((2 * np.linalg.multi_dot([J.toarray().T, w, res]), np.linalg.multi_dot([res.T, w, dev3, w, res]) - (w).dot(dev3).trace(), np.linalg.multi_dot([res.T, w, dev2, w, res]) - (w).dot(dev2).trace(), np.linalg.multi_dot([res.T, w, dev1, w, res]) - (w).dot(dev1).trace(), np.linalg.multi_dot([res.T, w, dev0, w, res]) - (w).dot(dev0).trace()))

                return H_l, resul


        H_l, resul = construction(H = H, J = J, n_gamma = n_gamma, g = g, reml = reml)
        var = np.linalg.inv(H_l.toarray()).dot(resul)
        dp = var[:p.shape[0]]
        dg = var[p.shape[0]:]

        
        # Here are the two ways of implementing the line search : the first one, only on the light cures parameters & the second on both models parameters.
        # One can comment or decomment the wanted fonction phi and crit.
        
        def phi(p):
            return ((w * ( y - model(p, jac = False)))**2).sum()

        def crit(t):
            return phi(p + t * dp)

        vect = np.hstack((p,g))
        """
        def phi(vect, n = len(p)):
            p = vect[:n]
            g = vect[n:]
            val = model(p)
            V = var_model(val, g)
            variance = np.diag(V) #+ np.diag(sig**2)
            w = variance
            return ((w * ( y - val))**2).sum()

        def crit(t):
            return phi(vect + t * var)
        """
        t, fval, ni, funcalls = scipy.optimize.brent(crit, brack=(0,1), full_output=True)

        p = p + t * dp
        g = g + t * dg
        print(p, dp, g, dg)
        v = model(p, jac=False)
        w = var_model(g)                
        res = w * (y- v)

        new_chi2 = (np.log(2 * np.pi) * len(res)  + (res**2).sum())  / (len(y) - (p.shape[0] + g.shape[0])) #+ np.log(np.linalg.det(w))
            
        dchi2 = new_chi2 - chi2
        logging.info(' red chi2: %f -> %f [dchi2=%f]' % (chi2, new_chi2, dchi2))

        chi2 = new_chi2
        if niter >= 1 :
            if dchi2 > 0.:
                logging.info('increasing chi2 !')
                niter += 1
                continue #break

            if abs(dchi2) <= dchi2_min:
                logging.info('dchi2 < dchi2_min, exiting')
                niter += 1
                break#continue

            if niter > max_iter:
                break

        niter += 1

    if plot :
        pl.figure()
        pl.plot(x,y, 'k.')
        
        pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(variance), ls = '--', color='r', linewidth = 2, label = 'real')
        if n_gamma == 1:
            pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(g[0]), ls = '-.', color='b', label='simu')
            pl.text(6,11, ' n_iter : %i (dchi< %f) \n\n p0 : %f, p1: %f \n g0 : %f' % (niter,dchi2_min,p[0],p[1],g[0]), fontsize= 15)
        if n_gamma == 2:
            pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(g[0] * x + g[1]), ls = '-.', color='b', label='simu')
            pl.text(6,12, ' n_iter : %i (dchi< %f) \n\n p0 : %f, p1: %f \n g0 : %f, g1: %f' % (niter,dchi2_min,p[0],p[1],g[0],g[1]), fontsize= 15)
        if n_gamma == 4:
            pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(g[0] * x**3 + g[1] * x**2 + g[2] * x + g[3]), ls = '-.', color='b', label='simu')
            pl.text(6,12, ' n_iter : %i (dchi< %f) \n\n p0 : %f, p1: %f \n g0 : %f, g1: %f, g2 : %f, g3: %f' % (niter,dchi2_min,p[0],p[1],g[0],g[1],g[2],g[3]), fontsize= 15)
        pl.legend()
    return p, g, model, var_model



class WdepModel(object):
    """
    Model for a polynomial dependance of f(x) for the variance.
    """
    def __init__(self, x):
        self.xdata = x
        self.pars = self.para_init()

    def para_init(self):
        fp = [("gamma0", 1), ("gamma1", 1)]
        fp = FitParameters(fp)
        fp['gamma0'] = 0.015
        fp['gamma1'] = 0.0045
        return fp
    
    def __call__(self, pvar, pmod, J = None):
        """
        return the matrix of the variance on the diagonal.

        Inputs : - pvar = Array, parameters of the variance model.
                 - pmod = Array, parameters of the straight line model.

        Output : - variance model evaluate with teh new parameters
                 - if J is given : dV = coo_matrix, the jacobian of the variance model.
        """

        value = pmod[0]*self.xdata + pmod[1]
        if J == None:
            return pvar[0] * value + pvar[1]
        else :
            N = len(pvar) + len(pmod)
            n = len(value)

            # derive parametre de J:                                                                                                                                                     
            ii_J = J.col
            jj_J = J.row
            val_J = value[J.row] * J.data * pvar[0]

            # derive V_{0}:                                                                                                                                                              
            ii_v = (J.shape[1] ) * np.ones(value.shape)
            jj_v = range(n)
            val_v = value

            # derive V_{1}:
            ii_v1 = (J.shape[1] + 1) * np.ones(value.shape)
            jj_v1 = range(n)
            val_v1 = np.ones(value.shape)

            i = np.hstack((ii_J, ii_v, ii_v1))
            j = np.hstack((jj_J, jj_v, jj_v1))
            val_jac = np.hstack((val_J, val_v, val_v1))

            dV = coo_matrix((val_jac, (i,j)), shape = (N, n))
            return pvar[0] * value + pvar[1], dV




def hessian(x, w, p, g, J, dV):
    """                                                                                                                                                                                                       
    Create Hessian matrix needed to minimized the chi2, in order to fit models of straight line and variances on data, using Newton-Raphson's method.                                                             
    Appearing on the left hand side of the equation.

    Input : - x = Array, is defition interval of the function. 
            - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(x) * len(x)).
            - p = Array, is the parameter of the STRAIGHT LINE model.
            - g = Array, is the parameter of the VARIANCE model.
            - J = coo_matrix, is the Jacobian of the straight line model.
            - dV = coo_matrix, is the Jacobian of the variance model.

    Output : - H = coo_matrix, sparse hessian matrix.
    """
    H0 = 2 * J.T * w * J
    H = coo_matrix(H0)
    H00 = H.data
    i00 = H.row
    j00 = H.col

    def trace(i, j, w = w, dV = dV):
        dV = dV.toarray()
        dVi = coo_matrix(np.diag(dV[i]))
        dVj = coo_matrix(np.diag(dV[j]))
        return (w * dVi * w * dVj).toarray().trace() 

    npar = p.shape[0] + g.shape[0]
    Tr = np.array([trace(i,j) for i in range(npar) for j in range(i, npar)])
    ii = [i for i in range(npar) for j in range(i, npar)]
    jj = [j for i in range(npar) for j in range(i, npar)]
    #tr = coo_matrix((Tr,(ii,jj)),shape = (npar, npar))                                                                                                                                                    
    #H_l = tr + tr.T - np.diag(tr.toarray().diagonal())                                                                                                                                                    
    H_l = coo_matrix((np.hstack((Tr,Tr)), (np.hstack((ii,jj)), np.hstack((jj, ii)))), shape = ( npar, npar))
    H_l = coo_matrix((np.hstack((H_l.data, -np.diag(H_l.toarray())/2.)), (np.hstack((H_l.row, range(H_l.toarray().shape[0]))), np.hstack((H_l.col, range(H_l.toarray().shape[0]))))), shape = (npar,npar))
    H = coo_matrix((H00, (i00, j00)), shape = ( npar, npar))
    return H + H_l

def grad(res, w, J, dV, p, g):
    """
    Create gradient vector needed to minimized the chi2, in order to fit models of straight line and variances on data, using Newton-Raphson's method.
    Appearing on the right hand side of the equation.

    Input : - res = Array, is the residual of the measurement substracted by the model.
            - w = Array, is the matrix in which each diagonal element is the inverse variance of the measurement ( size : len(res) * len(res)).
            - J = coo_matrix, is the Jacobian of the linear model.                                                                                                                                            
            - dV = coo_matrix, is the Jacobian of the variance model.
            - p = Array, is the parameter of the straight line model.
            - g = Array, is the parameter of the VARIANCE model.

    Output : - (-1) * resul = Array, the needed gradient vector.
    """
    def grad_theta(i, J = J, res = res, dV = dV, w = w):
        """
        Calculate the term i of the gradient vector.

        Input : - i = int, represent the i-th parameter term of the gradient.
        
        Output : - Float, the ith term
        """
        grad = (w.toarray().dot(np.diag(dV.toarray()[i]))).trace() - np.linalg.multi_dot([res.T, w.toarray(), np.diag(dV.toarray()[i]), w.toarray(), res])
        if i < p.shape[0]:
            grad += - 2 * np.linalg.multi_dot([J.toarray().T[i], w.toarray(), res])
        return grad
    resul = np.array([grad_theta(i) for i in range(p.shape[0] + g.shape[0])])
    return -resul

        
def fit_lin_var(N, max_iter = 100, ll = 0.1, dchi2_min = 0.1, plot = True):
    """
    Minimization of the chi2.

    Inputs :- N = int, number of point for the function.  
            - max_iter = int, is the maximum number of iteration wanted.
            - ll = float, help to inverse the hessian matrix, by adding somthing one its diagonal.
            - dchi2_min = float, variation of chi2 desired to reach.
            - plot = bool, True will plot.

    Outputs : - p = Array, the final parameter of the linear model.
             - g  = Array, the final parameter of the variance model.
    """
    x, y, vari = gen_var(N = N, ng = 'dep')
    model = SModel(x)
    var_model = WdepModel(x)

    p = model.pars.free
    g = var_model.pars.free
    dp = np.zeros(p.shape)
    dg = np.zeros(g.shape)

    v, J = model(p, jac = True)
    V, dV = var_model(g, p, J=J)
    w = np.diag(1/V) #+ np.diag((1/vari))
    
    res = y - v
    chi2 = (np.linalg.multi_dot([res.T,w,res]) + len(res) * np.log(np.pi) + np.log(np.diag(w)).sum()) / (N - 4)
    Dchi = 0.
    niter = 0

    while niter < max_iter:
        v, J = model(p, jac = True)
        V, dV = var_model(g, p, J=J)
        w = coo_matrix(np.diag(1/V)) #+ np.diag((1/vari)))

        res = y - v
        H = hessian(x = x, w = w, p = p, g = g, J = J, dV = dV) + ll * diags(np.ones(p.shape[0]+ g.shape[0]))
        resul = grad(res = res, w = w, J = J, dV = dV, p = p, g = g)
        var = np.linalg.inv(H.toarray()).dot(resul)
        dp = var[:p.shape[0]]
        dg = var[p.shape[0]:]
        
        vect = np.hstack((p,g))

        # Here are the two ways of implementing the line search : the first one, only on the straight line parameters & the second on both models parameters.
        # One can comment or decomment the wanted fonction phi and crit.
        
        def phi(p):
            return ((w * ( y - model(p, jac = False)))**2).sum()

        def crit(t):
            return phi(p + t * dp)

        """
        def phi(vect, n = p.shape[0]):
            p = vect[:n]
            g = vect[n:]
            val = model(p, jac = False)
            
            V = var_model(g, p)
            #V = var_model(val, g)
            #variance = np.diag(V) #+ np.diag(sig**2)
            #w = np.linalg.inv(variance)
            w = np.diag(1/V) 
            return ((w * ( y - val))**2).sum()

        def crit(t):
            return phi(vect + t * var)
        """

        t, fval, ni, funcalls = scipy.optimize.brent(crit, brack=(0,1), full_output=True)

        
        p = p + t * dp
        g = g + t * dg
        
        print('p', p, '\n', 'dp', dp, '\n', 'g', g, '\n', 'dg', dg,'\n', 't', t, '\n')

        v = model(p, jac=False)
        V = var_model(g, p)
        res = y - v
        w = np.diag(1/V) 
        new_chi2 = (np.linalg.multi_dot([res.T,w,res]) + len(res) * np.log(np.pi) + np.log(np.diag(w)).sum()) /	(N - 4)
        dchi2 = new_chi2 - chi2
        logging.info('iter %f : red chi2: %f -> %f [dchi2=%f]' % (niter, chi2, new_chi2, dchi2))

        chi2 = new_chi2
        if niter >= 1 :
            if dchi2 > 0.:
                logging.info('increasing chi2 !')
                niter += 1
                continue

            if abs(dchi2) <= dchi2_min:
                logging.info('dchi2 < dchi2_min, exiting')
                niter += 1
                break

            if niter > max_iter:
                break

        niter += 1
    if plot:
        pl.figure()
        pl.plot(x, y, color = 'k', marker = '.', ls ='')
        pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(vari), ls = '--', color='r', linewidth = 2, label='real')
        pl.errorbar(x, p[0] * x + p[1], yerr = np.sqrt(g[0] * (p[0] * x + p[1]) + g[1]), ls = '-.', color='b', label = 'simu')
        pl.title('N : %i' % (N))
        pl.legend()
        pl.text(6,12, ' n_iter : %i (dchi< %f) \n\n p0 : %f, p1: %f \n g0 : %f, g1: %f' % (niter,dchi2_min,p[0],p[1],g[0],g[1]), fontsize= 15)
    return p, g



def control(N, n_gamma, npts = 100, max_iter = 100, dchi2_min = 0.00001, dbin = 0.001, plot = True ):
    """                                                                                                                                                                                                           
    Control plot : histogram of the variance parameters for found for all the N realisations.

    Input : - N  = int, number of realisation of the fit.
            - n_gamma = 1, 2, 4 or 'dep', model of the variance.
            - npts = int, number of points.
            - max_iter = int, is the maximum number of iteration wanted.
            - dchi2_ min = dchi2_min = float, variation of chi2 desired to reach.
            - dbin = float, interval of histograms binning
            - plot = bool, True if we want to see each plot of the fit function.

    Output : - histogram
             - coef = array, parameters of each straight line model fitted.
             - gammas = array, parameters of each variance model fitted.
    """
    coef = np.zeros((N,2))
    if type(n_gamma) == int : 
        gammas = np.zeros((N, n_gamma))
        for i in range(N):
            x, y, var = gen_var(N = npts, ng = n_gamma)
            c, g, v, vm = fit_sqp_lin(y, x, var, n_gamma = n_gamma, max_iter = max_iter, dchi2_min = dchi2_min, plot = False)
            coef[i] = c
            gammas[i] = g

    elif n_gamma == 'dep':
        gammas = np.zeros((N, 2))
        for i in range(N):
            c, g = fit_lin_var(N = npts, max_iter = max_iter, ll = 0.1, dchi2_min = dchi2_min, plot = False)
            coef[i] = c
            gammas[i] = g

    a = 1.235
    b = 9.123

    if plot :
        pl.figure()
        xx, yy = np.histogram(coef.T[0])
        rg = (yy.min(), yy.max())
        nbins = int((rg[1] - rg[0])/ dbin)
        xh, yh, hh = pl.hist(coef.T[0], bins = nbins)#, range = (a-1,a+1))
        med = np.median(coef.T[0])
        moy = coef.T[0].mean()
        pl.plot(a*np.ones((10)), np.linspace(0, xh.max(), 10), 'r--')
        pl.plot(med * np.ones((10)), np.linspace(0, xh.max(), 10), 'b--')
        pl.plot(moy * np.ones((10)), np.linspace(0, xh.max(), 10), 'k--')
        pl.title('%i simu  de %i npts: a = %f ( => a_simu moy = %f med = %f)' % (N, npts, a, moy, med)) #coef.T[0].mean()))

        pl.figure()
        xx, yy = np.histogram(coef.T[1])
        rg = (yy.min(), yy.max())
        nbins = int((rg[1] - rg[0])/ dbin)
        xh, yh, hh = pl.hist(coef.T[1])#, bins = nbins)#, range = (b-1,b+1))
        med = np.median(coef.T[1])
        moy = coef.T[1].mean()
        pl.plot(b*np.ones((10)), np.linspace(0, xh.max(), 10), 'r--')
        pl.plot(med*np.ones((10)), np.linspace(0, xh.max(), 10), 'b--')
        pl.plot(moy*np.ones((10)), np.linspace(0, xh.max(), 10), 'k--')
        pl.title('%i simu  de %i npts : b = %f ( => b_simu moy = %f med = %f)' % (N, npts, b, moy, med))

        if n_gamma == 1:
            realg = np.array([1.356])
        elif n_gamma == 2:
            realg = np.array([1.325, 2.2222])
        elif n_gamma == 4:
            realg = 1/30. * np.array([1., -19., 133., 229.])
        elif n_gamma == 'dep':
            realg = np.array([0.01, 0.005])
        
        for n in range(len(realg)):
            pl.figure()

            xx, yy = np.histogram(gammas.T[n])
            rg = (yy.min(), yy.max())
            nbins = int((rg[1] - rg[0])/ dbin)
            
            xh, yh, hh = pl.hist(gammas.T[n], bins = nbins, range = rg)
            med = np.median(gammas.T[n])
            moy = gammas.T[n].mean()

            pl.plot(realg[n] * np.ones((10)), np.linspace(0, xh.max(), 10), 'r--')
            pl.plot(med*np.ones((10)), np.linspace(0, xh.max(), 10), 'b--')        
            pl.plot(gammas.T[n].mean()*np.ones((10)), np.linspace(0, xh.max(), 10), 'k--')
            pl.title('%i simu  de %i npts : gamma_%i = %f ( => gamma_%i_simu moy = %f med = %f)' % (N, npts, n, realg[n], n, moy, med))

    return coef, gammas


def control_moy(n_gamma, N = 500, nb_pts  = [50, 100, 250, 500, 1000], max_iter = 10, dbin = 0.01, plot = True ):
    """
    plot of the mean and mediane of all parameter for the "n_gamma" variance model for differents initial number of points.
    
    Inputs : - n_gamma = 1, 2, 4 or 'dep', variance model code 
             - N = int, number of realisations
             - nb_pts = list of int, number of initiale points for the function.
             - max_iter = int, is the maximum number of iteration wanted.
             - dbin = float, interval of histograms binning.
             - plot = bool, True if we want to see each plot of the control function.

    Outputs : plots                                                                                   
    """
    A = np.zeros((len(nb_pts),1))
    B = np.zeros((len(nb_pts),1))

    MA = np.zeros((len(nb_pts),1))
    MB = np.zeros((len(nb_pts),1))
    
    SA = np.zeros((len(nb_pts),1))
    SB = np.zeros((len(nb_pts),1))
    if type(n_gamma) != int:
        n = 2
    else :
        n = n_gamma
    G = np.zeros((len(nb_pts),n))
    MG = np.zeros((len(nb_pts),n))
    SG = np.zeros((len(nb_pts),n))
    for nn in range(len(nb_pts)):
        c, g = control(N = N, max_iter = max_iter, n_gamma = n_gamma, npts = nb_pts[nn], dbin = dbin, plot = plot)
        A[nn] = c.T[0].mean()
        B[nn] = c.T[1].mean()
        MA[nn] = np.median(c.T[0])
        MB[nn] = np.median(c.T[1])

        SA[nn] = c.T[0].std()
        SB[nn] = c.T[1].std()
        for i in range(n):
            G[nn][i] = g.T[i].mean()
            MG[nn][i] = np.median(g.T[i])
            SG[nn][i] = g.T[i].std()
    a = 1.235
    b = 9.123

    if n_gamma == 1:
        realg = np.array([1.356])
    elif n_gamma == 2:
        realg = np.array([1.325, 2.2222])
    elif n_gamma == 4:
        realg = 1/30. * np.array([1., -19., 133., 229.])
    else:
        realg = np.array([0.01, 0.005])

    pl.figure()
    for ii in range(len(nb_pts)):
        #pl.plot(A[ii],nb_pts[ii], '.', label ='%i pts' %(nb_pts[ii]))
        pl.errorbar(nb_pts[ii], A[ii], yerr = SA[ii], fmt='.', label ='moy %i pts' %(nb_pts[ii]))
        pl.errorbar(nb_pts[ii], MA[ii], yerr = SA[ii], fmt='*', label ='med %i pts' %(nb_pts[ii]))
    pl.plot(np.linspace(0,1000,10), a*np.ones(10), 'r--', label= 'real')
    #pl.plot(a*np.ones(10), np.linspace(0,1000,10), 'r--', label= 'real')
    pl.title('a')
    pl.legend()
    
    pl.figure()
    for ii in range(len(nb_pts)):
        pl.errorbar(nb_pts[ii], B[ii], yerr = SB[ii], fmt = '.', label ='moy %i pts' %(nb_pts[ii]))
        pl.errorbar(nb_pts[ii], MB[ii], yerr = SB[ii], fmt = '*', label ='med %i pts' %(nb_pts[ii]))
    pl.plot(np.linspace(0,1000,10), b * np.ones(10),'r--',label= 'real')
    pl.title('b')
    pl.legend()
    
    for iii in range(n):
        pl.figure()
        for ii in range(len(nb_pts)):
            pl.errorbar(nb_pts[ii], G[ii][iii], yerr = SG[ii][iii],fmt = '.', label ='moy %i pts' %(nb_pts[ii]))
            pl.errorbar(nb_pts[ii], MG[ii][iii], yerr = SG[ii][iii],fmt = '*', label ='med %i pts' %(nb_pts[ii]))
        pl.plot(np.linspace(0,1000,10), realg[iii] * np.ones(10), '--', label= 'real')
        pl.title( 'gamma_%i' % (iii))
        pl.legend()



import numpy as np
import scipy.sparse as sparse
import pandas
from matplotlib import pyplot as pl

from nacl.loglikelihood import LogLikelihood
from nacl.minimize import Minimizer
from nacl.models.toy1d import spline


def main_no_adapted_regularization(n=100, bsize=25):
    """
    """
    gen = spline.ToyDataGenerator(n=n, bsize=bsize, error_pedestal=0., delta=5.)
    d = gen.generate()
    reg = spline.Regularization(block_name='theta', mu0=1., mu2=1.)

    model = spline.Model(gen.x, gen.basis.grid)
    ll = LogLikelihood(model, reg=[reg], data=gen.get_data())
    v, grad, hess = ll(ll.pars.free, deriv=1)

    minz = Minimizer(ll)
    p_init = ll.pars.copy()
    p_init['theta'].full[:] = np.random.rand(len(p_init['theta'].full))
    minz.minimize_lm(p_init=p_init.free, lamb=1.E-6)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data)


def main_adapted_regularization(n=100, bsize=30, mu=1000., mu_model=1.E-6):
    """
    """
    gen = spline.ToyDataGenerator(n=n, bsize=bsize, error_pedestal=0., delta=5.)
    d = gen.generate()
    reg = spline.Regularization(block_name='theta', mu0=1., mu2=1.)

    model = spline.Model(gen.x, gen.basis.grid)
    ll = LogLikelihood(model, reg=[reg], data=gen.get_data())
    v, grad, hess = ll(ll.pars.free, deriv=1)

    minz = Minimizer(ll)
    p_init = ll.pars.copy()
    p_init['theta'].full[:] = np.random.rand(len(p_init['theta'].full))
    minz.minimize_lm(p_init=p_init.free, lamb=1.E-6)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data)

    _,J = model(ll.pars, jac=1)
    reg.adapt_regularization_strength(J, mu=mu, mu_model=mu_model)
    minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data)


def main_adapted_regularization_error_model(n=100, bsize=20,
                                            mu=1000, mu_model=1.E-6,
                                            error_pedestal=0.0075):
    """
    """
    gen = spline.ToyDataGenerator(n=n, bsize=bsize, error_pedestal=error_pedestal, delta=5.)
    d = gen.generate()
    reg = spline.Regularization(block_name='theta', mu0=2., mu2=1.)

    model = spline.Model(gen.x, gen.basis.grid)
    error_snake = spline.SimplePedestalErrorModel(model)
    ll = LogLikelihood(model, variance_model=error_snake, reg=[reg], data=gen.get_data())
    v, grad, hess = ll(ll.pars.free, deriv=1)

    minz = Minimizer(ll)
    p_init = ll.pars.copy()
    p_init['theta'].full[:] = np.random.rand(len(p_init['theta'].full))
    p_init['gamma'].full[:] = 1.
    r = minz.minimize_lm(p_init=p_init.free, lamb=1.E-6)
    print(r)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data)
    minz.plot()

    _,J = model(ll.pars, jac=1)
    reg.adapt_regularization_strength(J, mu=mu, mu_model=mu_model)
    r = minz.minimize_lm(p_init=ll.pars.free, lamb=1.E-6)
    spline.plot_model_and_data(minz.log_likelihood.model, minz.log_likelihood.pars, minz.log_likelihood.data, reg_lambda=reg.reg_lambda)
    print(r)
